# This code is part of Qiskit.
#
# (C) Copyright IBM 2022, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
High-level-synthesis transpiler pass.
"""

from __future__ import annotations

import typing
from functools import partial
from collections.abc import Callable

import numpy as np

from qiskit.circuit.annotated_operation import Modifier
from qiskit.circuit.operation import Operation
from qiskit.circuit.instruction import Instruction
from qiskit.converters import circuit_to_dag, dag_to_circuit
from qiskit.transpiler.basepasses import TransformationPass
from qiskit.circuit.quantumcircuit import QuantumCircuit
from qiskit.circuit import ControlledGate, EquivalenceLibrary, equivalence, Qubit
from qiskit.transpiler.passes.utils import control_flow
from qiskit.transpiler.target import Target
from qiskit.transpiler.coupling import CouplingMap
from qiskit.dagcircuit.dagcircuit import DAGCircuit
from qiskit.transpiler.exceptions import TranspilerError

from qiskit.circuit.annotated_operation import (
    AnnotatedOperation,
    InverseModifier,
    ControlModifier,
    PowerModifier,
)

from qiskit._accelerate.high_level_synthesis import QubitTracker, QubitContext
from .plugin import HighLevelSynthesisPluginManager

if typing.TYPE_CHECKING:
    from qiskit.dagcircuit import DAGOpNode


class HLSConfig:
    """The high-level-synthesis config allows to specify a list of "methods" used by
    :class:`~.HighLevelSynthesis` transformation pass to synthesize different types
    of higher-level objects.

    A higher-level object is an object of type :class:`~.Operation` (e.g., :class:`.Clifford` or
    :class:`.LinearFunction`).  Each object is referred to by its :attr:`~.Operation.name` field
    (e.g., ``"clifford"`` for :class:`.Clifford` objects), and the applicable synthesis methods are
    tied to this name.

    In the config, each method is specified in one of several ways:

    1. a tuple consisting of the name of a known synthesis plugin and a dictionary providing
       additional arguments for the algorithm.
    2. a tuple consisting of an instance of :class:`.HighLevelSynthesisPlugin` and additional
       arguments for the algorithm.
    3. a single string of a known synthesis plugin
    4. a single instance of :class:`.HighLevelSynthesisPlugin`.

    The following example illustrates different ways how a config file can be created::

        from qiskit.transpiler.passes.synthesis.high_level_synthesis import HLSConfig
        from qiskit.transpiler.passes.synthesis.high_level_synthesis import ACGSynthesisPermutation

        # All the ways to specify hls_config are equivalent
        hls_config = HLSConfig(permutation=[("acg", {})])
        hls_config = HLSConfig(permutation=["acg"])
        hls_config = HLSConfig(permutation=[(ACGSynthesisPermutation(), {})])
        hls_config = HLSConfig(permutation=[ACGSynthesisPermutation()])

    The names of the synthesis plugins should be declared in ``entry-points`` table for
    ``qiskit.synthesis`` in ``pyproject.toml``, in the form
    <higher-level-object-name>.<synthesis-method-name>.

    The standard higher-level-objects are recommended to have a synthesis method
    called "default", which would be called automatically when synthesizing these objects,
    without having to explicitly set these methods in the config.

    To avoid synthesizing a given higher-level-object, one can give it an empty list of methods.

    For an explicit example of using such config files, refer to the documentation for
    :class:`~.HighLevelSynthesis`.

    For an overview of the complete process of using high-level synthesis, see
    :ref:`using-high-level-synthesis-plugins`.
    """

    def __init__(
        self,
        use_default_on_unspecified: bool = True,
        plugin_selection: str = "sequential",
        plugin_evaluation_fn: Callable[[QuantumCircuit], int] | None = None,
        **kwargs,
    ):
        """Creates a high-level-synthesis config.

        Args:
            use_default_on_unspecified: if True, every higher-level-object without an
                explicitly specified list of methods will be synthesized using the "default"
                algorithm if it exists.
            plugin_selection: if set to ``"sequential"`` (default), for every higher-level-object
                the synthesis pass will consider the specified methods sequentially, stopping
                at the first method that is able to synthesize the object. If set to ``"all"``,
                all the specified methods will be considered, and the best synthesized circuit,
                according to ``plugin_evaluation_fn`` will be chosen.
            plugin_evaluation_fn: a callable that evaluates the quality of the synthesized
                quantum circuit; a smaller value means a better circuit. If ``None``, the
                quality of the circuit its size (i.e. the number of gates that it contains).
            kwargs: a dictionary mapping higher-level-objects to lists of synthesis methods.
        """
        self.use_default_on_unspecified = use_default_on_unspecified
        self.plugin_selection = plugin_selection
        self.plugin_evaluation_fn = (
            plugin_evaluation_fn if plugin_evaluation_fn is not None else lambda qc: qc.size()
        )
        self.methods = {}

        for key, value in kwargs.items():
            self.set_methods(key, value)

    def set_methods(self, hls_name, hls_methods):
        """Sets the list of synthesis methods for a given higher-level-object. This overwrites
        the lists of methods if also set previously."""
        self.methods[hls_name] = hls_methods


class HighLevelSynthesis(TransformationPass):
    r"""Synthesize higher-level objects and unroll custom definitions.

    The input to this pass is a DAG that may contain higher-level objects,
    including abstract mathematical objects (e.g., objects of type :class:`.LinearFunction`),
    annotated operations (objects of type :class:`.AnnotatedOperation`), and
    custom gates.

    In the most common use-case when either ``basis_gates`` or ``target`` is specified,
    all higher-level objects are synthesized, so the output is a :class:`.DAGCircuit`
    without such objects.
    More precisely, every gate in the output DAG is either directly supported by the target,
    or is in ``equivalence_library``.

    The abstract mathematical objects are synthesized using synthesis plugins, applying
    synthesis methods specified in the high-level-synthesis config (refer to the documentation
    for :class:`~.HLSConfig`).

    As an example, let us assume that ``op_a`` and ``op_b`` are names of two higher-level objects,
    that ``op_a``-objects have two synthesis methods ``default`` which does require any additional
    parameters and ``other`` with two optional integer parameters ``option_1`` and ``option_2``,
    that ``op_b``-objects have a single synthesis method ``default``, and ``qc`` is a quantum
    circuit containing ``op_a`` and ``op_b`` objects. The following code snippet::

        hls_config = HLSConfig(op_b=[("other", {"option_1": 7, "option_2": 4})])
        pm = PassManager([HighLevelSynthesis(hls_config=hls_config)])
        transpiled_qc = pm.run(qc)

    shows how to run the alternative synthesis method ``other`` for ``op_b``-objects, while using the
    ``default`` methods for all other high-level objects, including ``op_a``-objects.

    The annotated operations (consisting of a base operation and a list of inverse, control and power
    modifiers) are synthesizing recursively, first synthesizing the base operation, and then applying
    synthesis methods for creating inverted, controlled, or powered versions of that).

    The custom gates are synthesized by recursively unrolling their definitions, until every gate
    is either supported by the target or is in the equivalence library.

    When neither ``basis_gates`` nor ``target`` is specified, the pass synthesizes only the top-level
    abstract mathematical objects and annotated operations, without descending into the gate
    ``definitions``. This is consistent with the older behavior of the pass, allowing to synthesize
    some higher-level objects using plugins and leaving the other gates untouched.

    The high-level-synthesis passes information about available auxiliary qubits, and whether their
    state is clean (defined as :math:`|0\rangle`) or dirty (unknown state) to the synthesis routine
    via the respective arguments ``"num_clean_ancillas"`` and ``"num_dirty_ancillas"``.
    If ``qubits_initially_zero`` is ``True`` (default), the qubits are assumed to be in the
    :math:`|0\rangle` state. When appending a synthesized block using auxiliary qubits onto the
    circuit, we first use the clean auxiliary qubits.

    .. note::

        Synthesis methods are assumed to maintain the state of the auxiliary qubits.
        Concretely this means that clean auxiliary qubits must still be in the :math:`|0\rangle`
        state after the synthesized block, while dirty auxiliary qubits are re-used only
        as dirty qubits.

    """

    def __init__(
        self,
        hls_config: HLSConfig | None = None,
        coupling_map: CouplingMap | None = None,
        target: Target | None = None,
        use_qubit_indices: bool = False,
        equivalence_library: EquivalenceLibrary | None = None,
        basis_gates: list[str] | None = None,
        min_qubits: int = 0,
        qubits_initially_zero: bool = True,
    ):
        r"""
        HighLevelSynthesis initializer.

        Args:
            hls_config: Optional, the high-level-synthesis config that specifies synthesis methods
                and parameters for various high-level-objects in the circuit. If it is not specified,
                the default synthesis methods and parameters will be used.
            coupling_map: Optional, directed graph represented as a coupling map.
            target: Optional, the backend target to use for this pass. If it is specified,
                it will be used instead of the coupling map.
            use_qubit_indices: a flag indicating whether this synthesis pass is running before or after
                the layout is set, that is, whether the qubit indices of higher-level-objects correspond
                to qubit indices on the target backend.
            equivalence_library: The equivalence library used (instructions in this library will not
                be unrolled by this pass).
            basis_gates: Optional, target basis names to unroll to, e.g. `['u3', 'cx']`.
                Ignored if ``target`` is also specified.
            min_qubits: The minimum number of qubits for operations in the input
                dag to translate.
            qubits_initially_zero: Indicates whether the qubits are initially in the state
                :math:`|0\rangle`. This allows the high-level-synthesis to use clean auxiliary qubits
                (i.e. in the zero state) to synthesize an operation.
        """
        super().__init__()

        if hls_config is not None:
            self.hls_config = hls_config
        else:
            # When the config file is not provided, we will use the "default" method
            # to synthesize Operations (when available).
            self.hls_config = HLSConfig(True)

        self.hls_plugin_manager = HighLevelSynthesisPluginManager()
        self._coupling_map = coupling_map
        self._target = target
        self._use_qubit_indices = use_qubit_indices
        self.qubits_initially_zero = qubits_initially_zero
        if target is not None:
            self._coupling_map = self._target.build_coupling_map()
        self._equiv_lib = equivalence_library
        self._basis_gates = basis_gates
        self._min_qubits = min_qubits

        self._top_level_only = self._basis_gates is None and self._target is None

        # include path for when target exists but target.num_qubits is None (BasicSimulator)
        if not self._top_level_only and (self._target is None or self._target.num_qubits is None):
            basic_insts = {"measure", "reset", "barrier", "snapshot", "delay", "store"}
            self._device_insts = basic_insts | set(self._basis_gates)
        else:
            self._device_insts = set()

    def run(self, dag: DAGCircuit) -> DAGCircuit:
        """Run the HighLevelSynthesis pass on `dag`.

        Args:
            dag: input dag.

        Returns:
            Output dag with higher-level operations synthesized.

        Raises:
            TranspilerError: when the transpiler is unable to synthesize the given DAG
            (for instance, when the specified synthesis method is not available).
        """
        qubits = tuple(dag.find_bit(q).index for q in dag.qubits)
        context = QubitContext(list(range(len(dag.qubits))))
        tracker = QubitTracker(num_qubits=dag.num_qubits())
        if self.qubits_initially_zero:
            tracker.set_clean(context.to_globals(qubits))

        out_dag = self._run(dag, tracker, context, use_ancillas=True, top_level=True)
        return out_dag

    def _run(
        self,
        dag: DAGCircuit,
        tracker: QubitTracker,
        context: QubitContext,
        use_ancillas: bool,
        top_level: bool,
    ) -> DAGCircuit:
        """
        The main recursive function that synthesizes a DAGCircuit.

        Input:
            dag: the DAG to be synthesized.
            tracker: the global tracker, tracking the state of original qubits.
            context: the correspondence between the dag's qubits and the global qubits.
            use_ancillas: if True, synthesis algorithms are allowed to use ancillas.
            top_level: specifies if this is the top-level of the recursion.

        The function returns the synthesized DAG.

        Note that by using the auxiliary qubits to synthesize operations present in the input DAG,
        the synthesized DAG may be defined over more qubits than the input DAG. In this case,
        the function update in-place the global qubits tracker and extends the local-to-global
        context.
        """

        if dag.num_qubits() != context.num_qubits():
            raise TranspilerError("HighLevelSynthesis internal error.")

        # STEP 1: Check if HighLevelSynthesis can be skipped altogether. This is only
        # done at the top-level since this does not update the global qubits tracker.
        if top_level:
            for node in dag.op_nodes():
                qubits = tuple(dag.find_bit(q).index for q in node.qargs)
                if not self._definitely_skip_node(node, qubits, dag):
                    break
            else:
                # The for-loop terminates without reaching the break statement
                if dag.num_qubits() != context.num_qubits():
                    raise TranspilerError("HighLevelSynthesis internal error.")
                return dag

        # STEP 2: Analyze the nodes in the DAG. For each node in the DAG that needs
        # to be synthesized, we recursively synthesize it and store the result. For
        # instance, the result of synthesizing a custom gate is a DAGCircuit corresponding
        # to the (recursively synthesized) gate's definition. When the result is a
        # DAG, we also store its context (the mapping of its qubits to global qubits).
        # In addition, we keep track of the qubit states using the (global) qubits tracker.
        #
        # Note: This is a first version of a potentially more elaborate approach to find
        # good operation/ancilla allocations. The current approach is greedy and just gives
        # all available ancilla qubits to the current operation ("the-first-takes-all" approach).
        # It does not distribute ancilla qubits between different operations present in the DAG.
        synthesized_nodes = {}

        for node in dag.topological_op_nodes():
            qubits = tuple(dag.find_bit(q).index for q in node.qargs)
            processed = False
            synthesized = None
            synthesized_context = None

            # Start by handling special operations. Other cases can also be
            # considered: swaps, automatically simplifying control gate (e.g. if
            # a control is 0).
            if node.op.name in ["id", "delay", "barrier"]:
                # tracker not updated, these are no-ops
                processed = True

            elif node.op.name == "reset":
                # reset qubits to 0
                tracker.set_clean(context.to_globals(qubits))
                processed = True

            # check if synthesis for the operation can be skipped
            elif self._definitely_skip_node(node, qubits, dag):
                tracker.set_dirty(context.to_globals(qubits))

            # next check control flow
            elif node.is_control_flow():
                inner_context = context.restrict(qubits)
                synthesized = control_flow.map_blocks(
                    partial(
                        self._run,
                        tracker=tracker,
                        context=inner_context,
                        use_ancillas=False,
                        top_level=False,
                    ),
                    node.op,
                )

            # now we are free to synthesize
            else:
                # This returns the synthesized operation and its context (when the result is
                # a DAG, it's the correspondence between its qubits and the global qubits).
                # Also note that the DAG may use auxiliary qubits. The qubits tracker and the
                # current DAG's context are updated in-place.
                synthesized, synthesized_context = self._synthesize_operation(
                    node.op, qubits, tracker, context, use_ancillas=use_ancillas
                )

            # If the synthesis changed the operation (i.e. it is not None), store the result.
            if synthesized is not None:
                synthesized_nodes[node._node_id] = (synthesized, synthesized_context)

            # If the synthesis did not change anything, just update the qubit tracker.
            elif not processed:
                tracker.set_dirty(context.to_globals(qubits))

        # We did not change anything just return the input.
        if len(synthesized_nodes) == 0:
            if dag.num_qubits() != context.num_qubits():
                raise TranspilerError("HighLevelSynthesis internal error.")
            return dag

        # STEP 3. We rebuild the DAG with new operations. Note that we could also
        # check if no operation changed in size and substitute in-place, but rebuilding is
        # generally as fast or faster, unless very few operations are changed.
        out = dag.copy_empty_like()
        num_additional_qubits = context.num_qubits() - out.num_qubits()

        if num_additional_qubits > 0:
            out.add_qubits([Qubit() for _ in range(num_additional_qubits)])

        index_to_qubit = dict(enumerate(out.qubits))
        outer_to_local = context.to_local_mapping()

        for node in dag.topological_op_nodes():

            if op_tuple := synthesized_nodes.get(node._node_id, None):
                op, op_context = op_tuple

                if isinstance(op, Operation):
                    out.apply_operation_back(op, node.qargs, node.cargs)
                    continue

                if isinstance(op, QuantumCircuit):
                    op = circuit_to_dag(op, copy_operations=False)

                inner_to_global = op_context.to_global_mapping()
                if isinstance(op, DAGCircuit):
                    qubit_map = {
                        q: index_to_qubit[outer_to_local[inner_to_global[i]]]
                        for (i, q) in enumerate(op.qubits)
                    }
                    clbit_map = dict(zip(op.clbits, node.cargs))

                    for sub_node in op.op_nodes():
                        out.apply_operation_back(
                            sub_node.op,
                            tuple(qubit_map[qarg] for qarg in sub_node.qargs),
                            tuple(clbit_map[carg] for carg in sub_node.cargs),
                        )
                    out.global_phase += op.global_phase

                else:
                    raise TranspilerError(f"Unexpected synthesized type: {type(op)}")
            else:
                out.apply_operation_back(node.op, node.qargs, node.cargs, check=False)

        if out.num_qubits() != context.num_qubits():
            raise TranspilerError("HighLevelSynthesis internal error.")

        return out

    def _synthesize_operation(
        self,
        operation: Operation,
        qubits: tuple[int],
        tracker: QubitTracker,
        context: QubitContext,
        use_ancillas: bool,
    ) -> tuple[QuantumCircuit | Operation | DAGCircuit | None, QubitContext | None]:
        """
        Synthesizes an operation. The function receives the qubits on which the operation
        is defined in the current DAG, the correspondence between the qubits of the current
        DAG and the global qubits and the global qubits tracker. The function returns the
        result of synthesizing the operation. The value of `None` means that the operation
        should remain as it is. When it's a circuit, we also return the context, i.e. the
        correspondence of its local qubits and the global qubits. The function changes
        in-place the tracker (state of the global qubits), the qubits (when the synthesized
        operation is defined over additional ancilla qubits), and the context (to keep track
        of where these ancilla qubits maps to).
        """

        synthesized_context = None

        # Try to synthesize the operation. We'll go through the following options:
        #  (1) Annotations: if the operator is annotated, synthesize the base operation
        #       and then apply the modifiers. Returns a circuit (e.g. applying a power)
        #       or operation (e.g adding control on an X gate).
        #  (2) High-level objects: try running the battery of high-level synthesis plugins (e.g.
        #       if the operation is a Clifford). Returns a circuit.
        #  (3) Unrolling custom definitions: try defining the operation if it is not yet
        #       in the set of supported instructions. Returns a circuit.
        #
        # If any of the above were triggered, we will recurse and go again through these steps
        # until no further change occurred. At this point, we convert circuits to DAGs (the final
        # possible return type). If there was no change, we just return ``None``.
        num_original_qubits = len(qubits)
        qubits = list(qubits)

        synthesized = None

        # Try synthesizing via AnnotatedOperation. This is faster than an isinstance check
        # but a bit less safe since someone could create operations with a ``modifiers`` attribute.
        if len(modifiers := getattr(operation, "modifiers", [])) > 0:
            # Note: the base operation must be synthesized without using potential control qubits
            # used in the modifiers.
            num_ctrl = sum(
                mod.num_ctrl_qubits for mod in modifiers if isinstance(mod, ControlModifier)
            )
            baseop_qubits = qubits[num_ctrl:]  # reminder: control qubits are the first ones

            # get qubits of base operation
            control_qubits = qubits[0:num_ctrl]

            # Do not allow access to control qubits
            tracker.disable(context.to_globals(control_qubits))
            synthesized_base_op, _ = self._synthesize_operation(
                operation.base_op,
                baseop_qubits,
                tracker,
                context,
                use_ancillas=use_ancillas,
            )

            if synthesized_base_op is None:
                synthesized_base_op = operation.base_op
            elif isinstance(synthesized_base_op, DAGCircuit):
                synthesized_base_op = dag_to_circuit(synthesized_base_op)

            # Handle the case that synthesizing the base operation introduced
            # additional qubits (e.g. the base operation is a circuit that includes
            # an MCX gate).
            if synthesized_base_op.num_qubits > len(baseop_qubits):
                global_aux_qubits = tracker.borrow(
                    synthesized_base_op.num_qubits - len(baseop_qubits),
                    context.to_globals(baseop_qubits),
                )
                global_to_local = context.to_local_mapping()
                for aq in global_aux_qubits:
                    if aq in global_to_local:
                        qubits.append(global_to_local[aq])
                    else:
                        new_local_qubit = context.add_qubit(aq)
                        qubits.append(new_local_qubit)
            # Restore access to control qubits.
            tracker.enable(context.to_globals(control_qubits))

            # This step currently does not introduce ancilla qubits.
            synthesized = self._apply_annotations(synthesized_base_op, operation.modifiers)

        # If it was no AnnotatedOperation, try synthesizing via HLS or by unrolling.
        else:
            # Try synthesis via HLS -- which will return ``None`` if unsuccessful.
            indices = qubits if self._use_qubit_indices else None
            if len(hls_methods := self._methods_to_try(operation.name)) > 0:
                if use_ancillas:
                    num_clean_available = tracker.num_clean(context.to_globals(qubits))
                    num_dirty_available = tracker.num_dirty(context.to_globals(qubits))
                else:
                    num_clean_available = 0
                    num_dirty_available = 0
                synthesized = self._synthesize_op_using_plugins(
                    hls_methods,
                    operation,
                    indices,
                    num_clean_available,
                    num_dirty_available,
                )

                # It may happen that the plugin synthesis method uses clean/dirty ancilla qubits
                if (synthesized is not None) and (synthesized.num_qubits > len(qubits)):
                    # need to borrow more qubits from tracker
                    global_aux_qubits = tracker.borrow(
                        synthesized.num_qubits - len(qubits), context.to_globals(qubits)
                    )
                    global_to_local = context.to_local_mapping()

                    for aq in global_aux_qubits:
                        if aq in global_to_local:
                            qubits.append(global_to_local[aq])
                        else:
                            new_local_qubit = context.add_qubit(aq)
                            qubits.append(new_local_qubit)

            # If HLS did not apply, or was unsuccessful, try unrolling custom definitions.
            if synthesized is None and not self._top_level_only:
                synthesized = self._get_custom_definition(operation, indices)

        if synthesized is None:
            # if we didn't synthesize, there was nothing to unroll
            # updating the tracker will be handled upstream
            pass

        # if it has been synthesized, recurse and finally store the decomposition
        elif isinstance(synthesized, Operation):
            resynthesized, resynthesized_context = self._synthesize_operation(
                synthesized, qubits, tracker, context, use_ancillas=use_ancillas
            )

            if resynthesized is not None:
                synthesized = resynthesized
            else:
                tracker.set_dirty(context.to_globals(qubits))
            if isinstance(resynthesized, DAGCircuit):
                synthesized_context = resynthesized_context

        elif isinstance(synthesized, QuantumCircuit):
            # Synthesized is a quantum circuit which we want to process recursively.
            # For example, it's the definition circuit of a custom gate
            # or a circuit obtained by calling a synthesis method on a high-level-object.
            # In the second case, synthesized may have more qubits than the original node.

            as_dag = circuit_to_dag(synthesized, copy_operations=False)
            inner_context = context.restrict(qubits)

            if as_dag.num_qubits() != inner_context.num_qubits():
                raise TranspilerError("HighLevelSynthesis internal error.")

            # We save the current state of the tracker to be able to return the ancilla
            # qubits to the current positions. Note that at this point we do not know
            # which ancilla qubits will be allocated.
            saved_tracker = tracker.copy()
            synthesized = self._run(
                as_dag, tracker, inner_context, use_ancillas=use_ancillas, top_level=False
            )
            synthesized_context = inner_context

            if (synthesized is not None) and (synthesized.num_qubits() > len(qubits)):
                # need to borrow more qubits from tracker
                global_aux_qubits = tracker.borrow(
                    synthesized.num_qubits() - len(qubits), context.to_globals(qubits)
                )
                global_to_local = context.to_local_mapping()

                for aq in global_aux_qubits:
                    if aq in global_to_local:
                        qubits.append(global_to_local[aq])
                    else:
                        new_local_qubit = context.add_qubit(aq)
                        qubits.append(new_local_qubit)

            if len(qubits) > num_original_qubits:
                tracker.replace_state(
                    saved_tracker, context.to_globals(qubits[num_original_qubits:])
                )

        else:
            raise TranspilerError(f"Unexpected synthesized type: {type(synthesized)}")

        if isinstance(synthesized, DAGCircuit) and synthesized_context is None:
            raise TranspilerError("HighLevelSynthesis internal error.")

        return synthesized, synthesized_context

    def _get_custom_definition(
        self, inst: Instruction, qubits: list[int] | None
    ) -> QuantumCircuit | None:
        # check if the operation is already supported natively
        if not (isinstance(inst, ControlledGate) and inst._open_ctrl):
            # include path for when target exists but target.num_qubits is None (BasicSimulator)
            inst_supported = self._instruction_supported(inst.name, qubits)
            if inst_supported or (self._equiv_lib is not None and self._equiv_lib.has_entry(inst)):
                return None  # we support this operation already

        # if not, try to get the definition
        try:
            definition = inst.definition
        except (TypeError, AttributeError) as err:
            raise TranspilerError(f"HighLevelSynthesis was unable to define {inst.name}.") from err

        if definition is None:
            raise TranspilerError(f"HighLevelSynthesis was unable to synthesize {inst}.")

        return definition

    def _methods_to_try(self, name: str):
        """Get a sequence of methods to try for a given op name."""
        if (methods := self.hls_config.methods.get(name)) is not None:
            # the operation's name appears in the user-provided config,
            # we use the list of methods provided by the user
            return methods
        if (
            self.hls_config.use_default_on_unspecified
            and "default" in self.hls_plugin_manager.method_names(name)
        ):
            # the operation's name does not appear in the user-specified config,
            # we use the "default" method when instructed to do so and the "default"
            # method is available
            return ["default"]
        return []

    def _synthesize_op_using_plugins(
        self,
        hls_methods: list,
        op: Operation,
        qubits: list[int] | None,
        num_clean_ancillas: int = 0,
        num_dirty_ancillas: int = 0,
    ) -> QuantumCircuit | None:
        """
        Attempts to synthesize op using plugin mechanism.

        The arguments ``num_clean_ancillas`` and ``num_dirty_ancillas`` specify
        the number of clean and dirty qubits available to synthesize the given
        operation. A synthesis method does not need to use these additional qubits.

        Returns either the synthesized circuit or None (which may occur
        when no synthesis methods is available or specified, or when there is
        an insufficient number of auxiliary qubits).
        """
        hls_plugin_manager = self.hls_plugin_manager

        best_decomposition = None
        best_score = np.inf

        for method in hls_methods:
            # There are two ways to specify a synthesis method. The more explicit
            # way is to specify it as a tuple consisting of a synthesis algorithm and a
            # list of additional arguments, e.g.,
            #   ("kms", {"all_mats": 1, "max_paths": 100, "orig_circuit": 0}), or
            #   ("pmh", {}).
            # When the list of additional arguments is empty, one can also specify
            # just the synthesis algorithm, e.g.,
            #   "pmh".
            if isinstance(method, tuple):
                plugin_specifier, plugin_args = method
            else:
                plugin_specifier = method
                plugin_args = {}

            # There are two ways to specify a synthesis algorithm being run,
            # either by name, e.g. "kms" (which then should be specified in entry_points),
            # or directly as a class inherited from HighLevelSynthesisPlugin (which then
            # does not need to be specified in entry_points).
            if isinstance(plugin_specifier, str):
                if plugin_specifier not in hls_plugin_manager.method_names(op.name):
                    raise TranspilerError(
                        f"Specified method: {plugin_specifier} not found in available "
                        f"plugins for {op.name}"
                    )
                plugin_method = hls_plugin_manager.method(op.name, plugin_specifier)
            else:
                plugin_method = plugin_specifier

            # Set the number of available clean and dirty auxiliary qubits via plugin args.
            plugin_args["num_clean_ancillas"] = num_clean_ancillas
            plugin_args["num_dirty_ancillas"] = num_dirty_ancillas

            decomposition = plugin_method.run(
                op,
                coupling_map=self._coupling_map,
                target=self._target,
                qubits=qubits,
                **plugin_args,
            )

            # The synthesis methods that are not suited for the given higher-level-object
            # will return None.
            if decomposition is not None:
                if self.hls_config.plugin_selection == "sequential":
                    # In the "sequential" mode the first successful decomposition is
                    # returned.
                    best_decomposition = decomposition
                    break

                # In the "run everything" mode we update the best decomposition
                # discovered
                current_score = self.hls_config.plugin_evaluation_fn(decomposition)
                if current_score < best_score:
                    best_decomposition = decomposition
                    best_score = current_score

        return best_decomposition

    def _apply_annotations(
        self, synthesized: Operation | QuantumCircuit, modifiers: list[Modifier]
    ) -> QuantumCircuit:
        """
        Recursively synthesizes annotated operations.
        Returns either the synthesized operation or None (which occurs when the operation
        is not an annotated operation).
        """
        for modifier in modifiers:
            if isinstance(modifier, InverseModifier):
                # Both QuantumCircuit and Gate have inverse method
                synthesized = synthesized.inverse()

            elif isinstance(modifier, ControlModifier):
                # Both QuantumCircuit and Gate have control method, however for circuits
                # it is more efficient to avoid constructing the controlled quantum circuit.
                if isinstance(synthesized, QuantumCircuit):
                    synthesized = synthesized.to_gate()

                synthesized = synthesized.control(
                    num_ctrl_qubits=modifier.num_ctrl_qubits,
                    label=None,
                    ctrl_state=modifier.ctrl_state,
                    annotated=False,
                )

                if isinstance(synthesized, AnnotatedOperation):
                    raise TranspilerError(
                        "HighLevelSynthesis failed to synthesize the control modifier."
                    )

            elif isinstance(modifier, PowerModifier):
                # QuantumCircuit has power method, and Gate needs to be converted
                # to a quantum circuit.
                if not isinstance(synthesized, QuantumCircuit):
                    synthesized = _instruction_to_circuit(synthesized)

                synthesized = synthesized.power(modifier.power)

            else:
                raise TranspilerError(f"Unknown modifier {modifier}.")

        return synthesized

    def _definitely_skip_node(
        self, node: DAGOpNode, qubits: tuple[int] | None, dag: DAGCircuit
    ) -> bool:
        """Fast-path determination of whether a node can certainly be skipped (i.e. nothing will
        attempt to synthesise it) without accessing its Python-space `Operation`.

        This is tightly coupled to `_recursively_handle_op`; it exists as a temporary measure to
        avoid Python-space `Operation` creation from a `DAGOpNode` if we wouldn't do anything to the
        node (which is _most_ nodes)."""

        if (
            dag._has_calibration_for(node)
            or len(node.qargs) < self._min_qubits
            or node.is_directive()
            or (self._instruction_supported(node.name, qubits) and not node.is_control_flow())
        ):
            return True

        return (
            # The fast path is just for Rust-space standard gates (which excludes
            # `AnnotatedOperation`).
            node.is_standard_gate()
            # We don't have the fast-path for controlled gates over 3 or more qubits.
            # However, we most probably want the fast-path for controlled 2-qubit gates
            # (such as CX, CZ, CY, CH, CRX, and so on), so "_definitely_skip_node" should
            # not immediately return False when encountering a controlled gate over 2 qubits.
            and not (node.is_controlled_gate() and node.num_qubits >= 3)
            # If there are plugins to try, they need to be tried.
            and not self._methods_to_try(node.name)
            # If all the above constraints hold, and it's already supported or the basis translator
            # can handle it, we'll leave it be.
            and (
                # This uses unfortunately private details of `EquivalenceLibrary`, but so does the
                # `BasisTranslator`, and this is supposed to just be temporary til this is moved
                # into Rust space.
                self._equiv_lib is not None
                and equivalence.Key(name=node.name, num_qubits=node.num_qubits)
                in self._equiv_lib.keys()
            )
        )

    def _instruction_supported(self, name: str, qubits: tuple[int] | None) -> bool:
        # include path for when target exists but target.num_qubits is None (BasicSimulator)
        if self._target is None or self._target.num_qubits is None:
            return name in self._device_insts
        return self._target.instruction_supported(operation_name=name, qargs=qubits)


def _instruction_to_circuit(inst: Instruction) -> QuantumCircuit:
    circuit = QuantumCircuit(inst.num_qubits, inst.num_clbits)
    circuit.append(inst, circuit.qubits, circuit.clbits)
    return circuit



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2018.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Transpiler testing"""

import io
from logging import StreamHandler, getLogger
import unittest.mock
import sys

from qiskit import QuantumRegister, QuantumCircuit
from qiskit.transpiler import PassManager, TranspilerError
from qiskit.passmanager import DoWhileController, ConditionalController
from test import QiskitTestCase  # pylint: disable=wrong-import-order
from ._dummy_passes import (
    PassA_TP_NR_NP,
    PassB_TP_RA_PA,
    PassC_TP_RA_PA,
    PassD_TP_NR_NP,
    PassE_AP_NR_NP,
    PassF_reduce_dag_property,
    PassJ_Bad_NoReturn,
    PassK_check_fixed_point_property,
    PassM_AP_NR_NP,
)


class SchedulerTestCase(QiskitTestCase):
    """Asserts for the scheduler."""

    def assertScheduler(self, circuit, passmanager, expected):
        """
        Run `transpile(circuit, passmanager)` and check
        if the passes run as expected.

        Args:
            circuit (QuantumCircuit): Circuit to transform via transpilation.
            passmanager (PassManager): pass manager instance for the transpilation process
            expected (list): List of things the passes are logging
        """
        logger = "LocalLogger"
        with self.assertLogs(logger, level="INFO") as cm:
            out = passmanager.run(circuit)
        self.assertIsInstance(out, QuantumCircuit)
        self.assertEqual([record.message for record in cm.records], expected)

    def assertSchedulerRaises(self, circuit, passmanager, expected, exception_type):
        """
        Run `transpile(circuit, passmanager)` and check
        if the passes run as expected until exception_type is raised.

        Args:
            circuit (QuantumCircuit): Circuit to transform via transpilation
            passmanager (PassManager): pass manager instance for the transpilation process
            expected (list): List of things the passes are logging
            exception_type (Exception): Exception that is expected to be raised.
        """
        logger = "LocalLogger"
        with self.assertLogs(logger, level="INFO") as cm:
            self.assertRaises(exception_type, passmanager.run, circuit)
        self.assertEqual([record.message for record in cm.records], expected)


class TestPassManagerInit(SchedulerTestCase):
    """The pass manager sets things at init time."""

    def test_passes(self):
        """A single chain of passes, with Requests and Preserves, at __init__ time"""
        circuit = QuantumCircuit(QuantumRegister(1))
        passmanager = PassManager(
            passes=[
                PassC_TP_RA_PA(),  # Request: PassA / Preserves: PassA
                PassB_TP_RA_PA(),  # Request: PassA / Preserves: PassA
                PassD_TP_NR_NP(argument1=[1, 2]),  # Requires: {}/ Preserves: {}
                PassB_TP_RA_PA(),
            ]
        )
        self.assertScheduler(
            circuit,
            passmanager,
            [
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassC_TP_RA_PA",
                "run transformation pass PassB_TP_RA_PA",
                "run transformation pass PassD_TP_NR_NP",
                "argument [1, 2]",
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassB_TP_RA_PA",
            ],
        )


class TestUseCases(SchedulerTestCase):
    """Combine passes in different ways and checks that passes are run
    in the right order."""

    def setUp(self):
        super().setUp()
        self.circuit = QuantumCircuit(QuantumRegister(1))
        self.passmanager = PassManager()

    def test_chain(self):
        """A single chain of passes, with Requires and Preserves."""
        self.passmanager.append(PassC_TP_RA_PA())  # Requires: PassA / Preserves: PassA
        self.passmanager.append(PassB_TP_RA_PA())  # Requires: PassA / Preserves: PassA
        self.passmanager.append(PassD_TP_NR_NP(argument1=[1, 2]))  # Requires: {}/ Preserves: {}
        self.passmanager.append(PassB_TP_RA_PA())
        self.assertScheduler(
            self.circuit,
            self.passmanager,
            [
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassC_TP_RA_PA",
                "run transformation pass PassB_TP_RA_PA",
                "run transformation pass PassD_TP_NR_NP",
                "argument [1, 2]",
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassB_TP_RA_PA",
            ],
        )

    def test_conditional_passes_true(self):
        """A pass set with a conditional parameter. The callable is True."""
        self.passmanager.append(PassE_AP_NR_NP(True))
        self.passmanager.append(
            ConditionalController(
                PassA_TP_NR_NP(), condition=lambda property_set: property_set["property"]
            )
        )
        self.assertScheduler(
            self.circuit,
            self.passmanager,
            [
                "run analysis pass PassE_AP_NR_NP",
                "set property as True",
                "run transformation pass PassA_TP_NR_NP",
            ],
        )

    def test_conditional_passes_true_fc(self):
        """A pass set with a conditional parameter (with FlowController). The callable is True."""
        self.passmanager.append(PassE_AP_NR_NP(True))

        self.passmanager.append(
            ConditionalController(
                [PassA_TP_NR_NP()], condition=lambda property_set: property_set["property"]
            )
        )
        self.assertScheduler(
            self.circuit,
            self.passmanager,
            [
                "run analysis pass PassE_AP_NR_NP",
                "set property as True",
                "run transformation pass PassA_TP_NR_NP",
            ],
        )

    def test_conditional_passes_false(self):
        """A pass set with a conditional parameter. The callable is False."""
        self.passmanager.append(PassE_AP_NR_NP(False))
        self.passmanager.append(
            ConditionalController(
                PassA_TP_NR_NP(), condition=lambda property_set: property_set["property"]
            )
        )
        self.assertScheduler(
            self.circuit,
            self.passmanager,
            ["run analysis pass PassE_AP_NR_NP", "set property as False"],
        )

    def test_conditional_and_loop(self):
        """Run a conditional first, then a loop."""
        self.passmanager.append(PassE_AP_NR_NP(True))
        self.passmanager.append(
            ConditionalController(
                DoWhileController(
                    [
                        PassK_check_fixed_point_property(),
                        PassA_TP_NR_NP(),
                        PassF_reduce_dag_property(),
                    ],
                    do_while=lambda property_set: not property_set["property_fixed_point"],
                ),
                condition=lambda property_set: property_set["property"],
            )
        )
        self.assertScheduler(
            self.circuit,
            self.passmanager,
            [
                "run analysis pass PassE_AP_NR_NP",
                "set property as True",
                "run analysis pass PassG_calculates_dag_property",
                "set property as 8 (from dag.property)",
                "run analysis pass PassK_check_fixed_point_property",
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 6",
                "run analysis pass PassG_calculates_dag_property",
                "set property as 6 (from dag.property)",
                "run analysis pass PassK_check_fixed_point_property",
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 5",
                "run analysis pass PassG_calculates_dag_property",
                "set property as 5 (from dag.property)",
                "run analysis pass PassK_check_fixed_point_property",
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 4",
                "run analysis pass PassG_calculates_dag_property",
                "set property as 4 (from dag.property)",
                "run analysis pass PassK_check_fixed_point_property",
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 3",
                "run analysis pass PassG_calculates_dag_property",
                "set property as 3 (from dag.property)",
                "run analysis pass PassK_check_fixed_point_property",
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 2",
                "run analysis pass PassG_calculates_dag_property",
                "set property as 2 (from dag.property)",
                "run analysis pass PassK_check_fixed_point_property",
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 2",
                "run analysis pass PassG_calculates_dag_property",
                "set property as 2 (from dag.property)",
                "run analysis pass PassK_check_fixed_point_property",
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 2",
            ],
        )

    def test_loop_and_conditional(self):
        """Run a loop first, then a conditional."""
        self.passmanager.append(PassK_check_fixed_point_property())
        self.passmanager.append(
            ConditionalController(
                DoWhileController(
                    [
                        PassK_check_fixed_point_property(),
                        PassA_TP_NR_NP(),
                        PassF_reduce_dag_property(),
                    ],
                    do_while=lambda property_set: not property_set["property_fixed_point"],
                ),
                condition=lambda property_set: not property_set["property_fixed_point"],
            )
        )
        self.assertScheduler(
            self.circuit,
            self.passmanager,
            [
                "run analysis pass PassG_calculates_dag_property",
                "set property as 8 (from dag.property)",
                "run analysis pass PassK_check_fixed_point_property",
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 6",
                "run analysis pass PassG_calculates_dag_property",
                "set property as 6 (from dag.property)",
                "run analysis pass PassK_check_fixed_point_property",
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 5",
                "run analysis pass PassG_calculates_dag_property",
                "set property as 5 (from dag.property)",
                "run analysis pass PassK_check_fixed_point_property",
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 4",
                "run analysis pass PassG_calculates_dag_property",
                "set property as 4 (from dag.property)",
                "run analysis pass PassK_check_fixed_point_property",
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 3",
                "run analysis pass PassG_calculates_dag_property",
                "set property as 3 (from dag.property)",
                "run analysis pass PassK_check_fixed_point_property",
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 2",
                "run analysis pass PassG_calculates_dag_property",
                "set property as 2 (from dag.property)",
                "run analysis pass PassK_check_fixed_point_property",
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 2",
                "run analysis pass PassG_calculates_dag_property",
                "set property as 2 (from dag.property)",
                "run analysis pass PassK_check_fixed_point_property",
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 2",
            ],
        )

    def test_do_not_repeat_based_on_preservation(self):
        """When a pass is still a valid pass (because the following passes
        preserved it), it should not run again."""
        self.passmanager.append([PassB_TP_RA_PA(), PassA_TP_NR_NP(), PassB_TP_RA_PA()])
        self.assertScheduler(
            self.circuit,
            self.passmanager,
            ["run transformation pass PassA_TP_NR_NP", "run transformation pass PassB_TP_RA_PA"],
        )

    def test_do_not_repeat_based_on_idempotence(self):
        """Repetition can be optimized to a single execution when
        the pass is idempotent."""
        self.passmanager.append(PassA_TP_NR_NP())
        self.passmanager.append([PassA_TP_NR_NP(), PassA_TP_NR_NP()])
        self.passmanager.append(PassA_TP_NR_NP())
        self.assertScheduler(
            self.circuit, self.passmanager, ["run transformation pass PassA_TP_NR_NP"]
        )

    def test_non_idempotent_pass(self):
        """Two or more runs of a non-idempotent pass cannot be optimized."""
        self.passmanager.append(PassF_reduce_dag_property())
        self.passmanager.append([PassF_reduce_dag_property(), PassF_reduce_dag_property()])
        self.passmanager.append(PassF_reduce_dag_property())
        self.assertScheduler(
            self.circuit,
            self.passmanager,
            [
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 6",
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 5",
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 4",
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 3",
            ],
        )

    def test_analysis_pass_is_idempotent(self):
        """Analysis passes are idempotent."""
        passmanager = PassManager()
        passmanager.append(PassE_AP_NR_NP(argument1=1))
        passmanager.append(PassE_AP_NR_NP(argument1=1))
        self.assertScheduler(
            self.circuit, passmanager, ["run analysis pass PassE_AP_NR_NP", "set property as 1"]
        )

    def test_ap_before_and_after_a_tp(self):
        """A default transformation does not preserves anything
        and analysis passes need to be re-run"""
        passmanager = PassManager()
        passmanager.append(PassE_AP_NR_NP(argument1=1))
        passmanager.append(PassA_TP_NR_NP())
        passmanager.append(PassE_AP_NR_NP(argument1=1))
        self.assertScheduler(
            self.circuit,
            passmanager,
            [
                "run analysis pass PassE_AP_NR_NP",
                "set property as 1",
                "run transformation pass PassA_TP_NR_NP",
                "run analysis pass PassE_AP_NR_NP",
                "set property as 1",
            ],
        )

    def test_pass_no_return(self):
        """Transformation passes that don't return a DAG raise error."""
        self.passmanager.append(PassJ_Bad_NoReturn())
        self.assertSchedulerRaises(
            self.circuit,
            self.passmanager,
            ["run transformation pass PassJ_Bad_NoReturn"],
            TranspilerError,
        )

    def test_fixed_point_pass(self):
        """A pass set with a do_while parameter that checks for a fixed point."""
        self.passmanager.append(
            DoWhileController(
                [
                    PassK_check_fixed_point_property(),
                    PassA_TP_NR_NP(),
                    PassF_reduce_dag_property(),
                ],
                do_while=lambda property_set: not property_set["property_fixed_point"],
            )
        )
        self.assertScheduler(
            self.circuit,
            self.passmanager,
            [
                "run analysis pass PassG_calculates_dag_property",
                "set property as 8 (from dag.property)",
                "run analysis pass PassK_check_fixed_point_property",
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 6",
                "run analysis pass PassG_calculates_dag_property",
                "set property as 6 (from dag.property)",
                "run analysis pass PassK_check_fixed_point_property",
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 5",
                "run analysis pass PassG_calculates_dag_property",
                "set property as 5 (from dag.property)",
                "run analysis pass PassK_check_fixed_point_property",
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 4",
                "run analysis pass PassG_calculates_dag_property",
                "set property as 4 (from dag.property)",
                "run analysis pass PassK_check_fixed_point_property",
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 3",
                "run analysis pass PassG_calculates_dag_property",
                "set property as 3 (from dag.property)",
                "run analysis pass PassK_check_fixed_point_property",
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 2",
                "run analysis pass PassG_calculates_dag_property",
                "set property as 2 (from dag.property)",
                "run analysis pass PassK_check_fixed_point_property",
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 2",
                "run analysis pass PassG_calculates_dag_property",
                "set property as 2 (from dag.property)",
                "run analysis pass PassK_check_fixed_point_property",
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 2",
            ],
        )

    def test_fixed_point_fc(self):
        """A fixed point scheduler with flow control."""
        self.passmanager.append(
            DoWhileController(
                [
                    PassK_check_fixed_point_property(),
                    PassA_TP_NR_NP(),
                    PassF_reduce_dag_property(),
                ],
                do_while=lambda property_set: not property_set["property_fixed_point"],
            )
        )

        expected = [
            "run analysis pass PassG_calculates_dag_property",
            "set property as 8 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 6",
            "run analysis pass PassG_calculates_dag_property",
            "set property as 6 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 5",
            "run analysis pass PassG_calculates_dag_property",
            "set property as 5 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 4",
            "run analysis pass PassG_calculates_dag_property",
            "set property as 4 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 3",
            "run analysis pass PassG_calculates_dag_property",
            "set property as 3 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 2",
            "run analysis pass PassG_calculates_dag_property",
            "set property as 2 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 2",
            "run analysis pass PassG_calculates_dag_property",
            "set property as 2 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 2",
        ]

        self.assertScheduler(self.circuit, self.passmanager, expected)

    def test_fixed_point_pass_max_iteration(self):
        """A pass set with a do_while parameter that checks that
        the max_iteration is raised."""
        self.passmanager.append(
            DoWhileController(
                [PassK_check_fixed_point_property(), PassA_TP_NR_NP(), PassF_reduce_dag_property()],
                do_while=lambda property_set: not property_set["property_fixed_point"],
                options={"max_iteration": 2},
            ),
        )
        self.assertSchedulerRaises(
            self.circuit,
            self.passmanager,
            [
                "run analysis pass PassG_calculates_dag_property",
                "set property as 8 (from dag.property)",
                "run analysis pass PassK_check_fixed_point_property",
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 6",
                "run analysis pass PassG_calculates_dag_property",
                "set property as 6 (from dag.property)",
                "run analysis pass PassK_check_fixed_point_property",
                "run transformation pass PassA_TP_NR_NP",
                "run transformation pass PassF_reduce_dag_property",
                "dag property = 5",
            ],
            TranspilerError,
        )

    def test_fresh_initial_state(self):
        """New construction gives fresh instance."""
        self.passmanager.append(PassM_AP_NR_NP(argument1=1))
        self.passmanager.append(PassA_TP_NR_NP())
        self.passmanager.append(PassM_AP_NR_NP(argument1=1))
        self.assertScheduler(
            self.circuit,
            self.passmanager,
            [
                "run analysis pass PassM_AP_NR_NP",
                "self.argument1 = 2",
                "run transformation pass PassA_TP_NR_NP",
                "run analysis pass PassM_AP_NR_NP",
                "self.argument1 = 2",
            ],
        )

    def test_nested_conditional_in_loop(self):
        """Run a loop with a nested conditional."""
        nested_conditional = ConditionalController(
            PassA_TP_NR_NP(), condition=lambda property_set: property_set["property"] >= 5
        )
        self.passmanager.append(
            DoWhileController(
                [
                    PassK_check_fixed_point_property(),
                    nested_conditional,
                    PassF_reduce_dag_property(),
                ],
                do_while=lambda property_set: not property_set["property_fixed_point"],
            )
        )
        expected = [
            "run analysis pass PassG_calculates_dag_property",
            "set property as 8 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 6",
            "run analysis pass PassG_calculates_dag_property",
            "set property as 6 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 5",
            "run analysis pass PassG_calculates_dag_property",
            "set property as 5 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 4",
            "run analysis pass PassG_calculates_dag_property",
            "set property as 4 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 3",
            "run analysis pass PassG_calculates_dag_property",
            "set property as 3 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 2",
            "run analysis pass PassG_calculates_dag_property",
            "set property as 2 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 2",
            "run analysis pass PassG_calculates_dag_property",
            "set property as 2 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 2",
        ]
        self.assertScheduler(self.circuit, self.passmanager, expected)


class TestControlFlowPlugin(SchedulerTestCase):
    """Testing the control flow plugin system."""

    def setUp(self):
        super().setUp()
        self.passmanager = PassManager()
        self.circuit = QuantumCircuit(QuantumRegister(1))


class TestDumpPasses(SchedulerTestCase):
    """Testing the passes method."""

    def test_passes(self):
        """Dump passes in different FlowControllerLinear"""
        passmanager = PassManager()
        passmanager.append(PassC_TP_RA_PA())
        passmanager.append(PassB_TP_RA_PA())
        expected = PassC_TP_RA_PA(), PassB_TP_RA_PA()
        self.assertEqual(expected, passmanager.to_flow_controller().tasks)

    def test_passes_in_linear(self):
        """Dump passes in the same FlowControllerLinear"""
        passmanager = PassManager(
            passes=[
                PassC_TP_RA_PA(),
                PassB_TP_RA_PA(),
                PassD_TP_NR_NP(argument1=[1, 2]),
                PassB_TP_RA_PA(),
            ]
        )
        expected = (
            PassC_TP_RA_PA(),
            PassB_TP_RA_PA(),
            PassD_TP_NR_NP(argument1=[1, 2]),
            PassB_TP_RA_PA(),
        )
        self.assertEqual(expected, passmanager.to_flow_controller().tasks)

    def test_conditional_and_loop(self):
        """Dump passes with a conditional and a loop."""
        nested_controller = ConditionalController(
            DoWhileController(
                [
                    PassK_check_fixed_point_property(),
                    PassA_TP_NR_NP(),
                    PassF_reduce_dag_property(),
                ],
                do_while=lambda property_set: not property_set["property_fixed_point"],
            ),
            condition=lambda property_set: property_set["property_fixed_point"],
        )

        passmanager = PassManager()
        passmanager.append(PassE_AP_NR_NP(True))
        passmanager.append(nested_controller)

        expected = (PassE_AP_NR_NP(True), nested_controller)
        self.assertEqual(expected, passmanager.to_flow_controller().tasks)


class StreamHandlerRaiseException(StreamHandler):
    """Handler class that will raise an exception on formatting errors."""

    def handleError(self, record):
        raise sys.exc_info()


class TestLogPasses(QiskitTestCase):
    """Testing the log_passes option."""

    def setUp(self):
        super().setUp()
        logger = getLogger()
        self.addCleanup(logger.setLevel, logger.level)
        logger.setLevel("DEBUG")
        self.output = io.StringIO()
        logger.addHandler(StreamHandlerRaiseException(self.output))
        self.circuit = QuantumCircuit(QuantumRegister(1))

    def assertPassLog(self, passmanager, list_of_passes):
        """Runs the passmanager and checks that the elements in
        passmanager.property_set['pass_log'] match list_of_passes (the names)."""
        passmanager.run(self.circuit)
        self.output.seek(0)
        # Filter unrelated log lines
        output_lines = self.output.readlines()
        pass_log_lines = [x for x in output_lines if x.startswith("Pass:")]
        for index, pass_name in enumerate(list_of_passes):
            self.assertTrue(pass_log_lines[index].startswith(f"Pass: {pass_name} -"))

    def test_passes(self):
        """Dump passes in different FlowControllerLinear"""
        passmanager = PassManager()
        passmanager.append(PassC_TP_RA_PA())
        passmanager.append(PassB_TP_RA_PA())

        self.assertPassLog(passmanager, ["PassA_TP_NR_NP", "PassC_TP_RA_PA", "PassB_TP_RA_PA"])

    def test_passes_in_linear(self):
        """Dump passes in the same FlowControllerLinear"""
        passmanager = PassManager(
            passes=[
                PassC_TP_RA_PA(),
                PassB_TP_RA_PA(),
                PassD_TP_NR_NP(argument1=[1, 2]),
                PassB_TP_RA_PA(),
            ]
        )

        self.assertPassLog(
            passmanager,
            [
                "PassA_TP_NR_NP",
                "PassC_TP_RA_PA",
                "PassB_TP_RA_PA",
                "PassD_TP_NR_NP",
                "PassA_TP_NR_NP",
                "PassB_TP_RA_PA",
            ],
        )

    def test_conditional_and_loop(self):
        """Dump passes with a conditional and a loop"""
        passmanager = PassManager()
        passmanager.append(PassE_AP_NR_NP(True))
        passmanager.append(
            ConditionalController(
                DoWhileController(
                    [
                        PassK_check_fixed_point_property(),
                        PassA_TP_NR_NP(),
                        PassF_reduce_dag_property(),
                    ],
                    do_while=lambda property_set: not property_set["property_fixed_point"],
                ),
                condition=lambda property_set: property_set["property_fixed_point"],
            )
        )
        self.assertPassLog(passmanager, ["PassE_AP_NR_NP"])


class TestPassManagerReuse(SchedulerTestCase):
    """The PassManager instance should be reusable."""

    def setUp(self):
        super().setUp()
        self.passmanager = PassManager()
        self.circuit = QuantumCircuit(QuantumRegister(1))

    def test_chain_twice(self):
        """Run a chain twice."""
        self.passmanager.append(PassC_TP_RA_PA())  # Request: PassA / Preserves: PassA
        self.passmanager.append(PassB_TP_RA_PA())  # Request: PassA / Preserves: PassA

        expected = [
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassC_TP_RA_PA",
            "run transformation pass PassB_TP_RA_PA",
        ]

        self.assertScheduler(self.circuit, self.passmanager, expected)
        self.assertScheduler(self.circuit, self.passmanager, expected)

    def test_conditional_twice(self):
        """Run a conditional twice."""
        self.passmanager.append(PassE_AP_NR_NP(True))
        self.passmanager.append(
            ConditionalController(
                PassA_TP_NR_NP(), condition=lambda property_set: property_set["property"]
            )
        )

        expected = [
            "run analysis pass PassE_AP_NR_NP",
            "set property as True",
            "run transformation pass PassA_TP_NR_NP",
        ]

        self.assertScheduler(self.circuit, self.passmanager, expected)
        self.assertScheduler(self.circuit, self.passmanager, expected)

    def test_fixed_point_twice(self):
        """A fixed point scheduler, twice."""
        self.passmanager.append(
            DoWhileController(
                [
                    PassK_check_fixed_point_property(),
                    PassA_TP_NR_NP(),
                    PassF_reduce_dag_property(),
                ],
                do_while=lambda property_set: not property_set["property_fixed_point"],
            )
        )

        expected = [
            "run analysis pass PassG_calculates_dag_property",
            "set property as 8 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 6",
            "run analysis pass PassG_calculates_dag_property",
            "set property as 6 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 5",
            "run analysis pass PassG_calculates_dag_property",
            "set property as 5 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 4",
            "run analysis pass PassG_calculates_dag_property",
            "set property as 4 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 3",
            "run analysis pass PassG_calculates_dag_property",
            "set property as 3 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 2",
            "run analysis pass PassG_calculates_dag_property",
            "set property as 2 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 2",
            "run analysis pass PassG_calculates_dag_property",
            "set property as 2 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 2",
        ]

        self.assertScheduler(self.circuit, self.passmanager, expected)
        self.assertScheduler(self.circuit, self.passmanager, expected)


class TestPassManagerChanges(SchedulerTestCase):
    """Test PassManager manipulation with changes"""

    def setUp(self):
        super().setUp()
        self.passmanager = PassManager()
        self.circuit = QuantumCircuit(QuantumRegister(1))

    def test_replace0(self):
        """Test passmanager.replace(0, ...)."""
        self.passmanager.append(PassC_TP_RA_PA())  # Request: PassA / Preserves: PassA
        self.passmanager.append(PassB_TP_RA_PA())  # Request: PassA / Preserves: PassA

        self.passmanager.replace(0, PassB_TP_RA_PA())

        expected = [
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassB_TP_RA_PA",
        ]
        self.assertScheduler(self.circuit, self.passmanager, expected)

    def test_replace1(self):
        """Test passmanager.replace(1, ...)."""
        self.passmanager.append(PassC_TP_RA_PA())  # Request: PassA / Preserves: PassA
        self.passmanager.append(PassB_TP_RA_PA())  # Request: PassA / Preserves: PassA

        self.passmanager.replace(1, PassC_TP_RA_PA())

        expected = [
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassC_TP_RA_PA",
        ]
        self.assertScheduler(self.circuit, self.passmanager, expected)

    def test_remove0(self):
        """Test passmanager.remove(0)."""
        self.passmanager.append(PassC_TP_RA_PA())  # Request: PassA / Preserves: PassA
        self.passmanager.append(PassB_TP_RA_PA())  # Request: PassA / Preserves: PassA

        self.passmanager.remove(0)

        expected = [
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassB_TP_RA_PA",
        ]
        self.assertScheduler(self.circuit, self.passmanager, expected)

    def test_remove1(self):
        """Test passmanager.remove(1)."""
        self.passmanager.append(PassC_TP_RA_PA())  # Request: PassA / Preserves: PassA
        self.passmanager.append(PassB_TP_RA_PA())  # Request: PassA / Preserves: PassA

        self.passmanager.remove(1)

        expected = [
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassC_TP_RA_PA",
        ]
        self.assertScheduler(self.circuit, self.passmanager, expected)

    def test_remove_minus_1(self):
        """Test passmanager.remove(-1)."""
        self.passmanager.append(PassA_TP_NR_NP())
        self.passmanager.append(PassB_TP_RA_PA())  # Request: PassA / Preserves: PassA

        self.passmanager.remove(-1)

        expected = ["run transformation pass PassA_TP_NR_NP"]
        self.assertScheduler(self.circuit, self.passmanager, expected)

    def test_setitem(self):
        """Test passmanager[1] = ..."""
        self.passmanager.append(PassC_TP_RA_PA())  # Request: PassA / Preserves: PassA
        self.passmanager.append(PassB_TP_RA_PA())  # Request: PassA / Preserves: PassA

        self.passmanager[1] = PassC_TP_RA_PA()

        expected = [
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassC_TP_RA_PA",
        ]
        self.assertScheduler(self.circuit, self.passmanager, expected)

    def test_replace_with_conditional(self):
        """Replace a pass with a conditional pass."""
        self.passmanager.append(PassE_AP_NR_NP(False))
        self.passmanager.append(PassB_TP_RA_PA())

        self.passmanager.replace(
            1,
            ConditionalController(
                PassA_TP_NR_NP(), condition=lambda property_set: property_set["property"]
            ),
        )

        expected = ["run analysis pass PassE_AP_NR_NP", "set property as False"]
        self.assertScheduler(self.circuit, self.passmanager, expected)

    def test_replace_error(self):
        """Replace a non-existing index."""
        self.passmanager.append(PassB_TP_RA_PA())

        with self.assertRaises(TranspilerError):
            self.passmanager.replace(99, PassA_TP_NR_NP())


class TestPassManagerSlicing(SchedulerTestCase):
    """test PassManager slicing."""

    def setUp(self):
        super().setUp()
        self.passmanager = PassManager()
        self.circuit = QuantumCircuit(QuantumRegister(1))

    def test_empty_passmanager_length(self):
        """test len(PassManager) when PassManager is empty"""
        length = len(self.passmanager)

        expected_length = 0
        self.assertEqual(length, expected_length)

    def test_passmanager_length(self):
        """test len(PassManager) when PassManager is not empty"""
        self.passmanager.append(PassA_TP_NR_NP())
        self.passmanager.append(PassA_TP_NR_NP())

        length = len(self.passmanager)

        expected_length = 2
        self.assertEqual(length, expected_length)

    def test_accessing_passmanager_by_index(self):
        """test accessing PassManager's passes by index"""
        self.passmanager.append(PassB_TP_RA_PA())
        self.passmanager.append(PassC_TP_RA_PA())

        new_passmanager = self.passmanager[1]

        expected = [
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassC_TP_RA_PA",
        ]
        self.assertScheduler(self.circuit, new_passmanager, expected)

    def test_accessing_passmanager_by_index_with_condition(self):
        """test accessing PassManager's conditioned passes by index"""
        self.passmanager.append(PassF_reduce_dag_property())
        self.passmanager.append(
            ConditionalController(
                DoWhileController(
                    [
                        PassK_check_fixed_point_property(),
                        PassA_TP_NR_NP(),
                        PassF_reduce_dag_property(),
                    ],
                    do_while=lambda property_set: not property_set["property_fixed_point"],
                ),
                condition=lambda property_set: True,
            )
        )

        new_passmanager = self.passmanager[1]

        expected = [
            "run analysis pass PassG_calculates_dag_property",
            "set property as 8 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 6",
            "run analysis pass PassG_calculates_dag_property",
            "set property as 6 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 5",
            "run analysis pass PassG_calculates_dag_property",
            "set property as 5 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 4",
            "run analysis pass PassG_calculates_dag_property",
            "set property as 4 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 3",
            "run analysis pass PassG_calculates_dag_property",
            "set property as 3 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 2",
            "run analysis pass PassG_calculates_dag_property",
            "set property as 2 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 2",
            "run analysis pass PassG_calculates_dag_property",
            "set property as 2 (from dag.property)",
            "run analysis pass PassK_check_fixed_point_property",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassF_reduce_dag_property",
            "dag property = 2",
        ]
        self.assertScheduler(self.circuit, new_passmanager, expected)

    def test_accessing_passmanager_by_range(self):
        """test accessing PassManager's passes by range"""
        self.passmanager.append(PassC_TP_RA_PA())
        self.passmanager.append(PassB_TP_RA_PA())
        self.passmanager.append(PassC_TP_RA_PA())
        self.passmanager.append(PassD_TP_NR_NP())

        new_passmanager = self.passmanager[1:3]

        expected = [
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassB_TP_RA_PA",
            "run transformation pass PassC_TP_RA_PA",
        ]
        self.assertScheduler(self.circuit, new_passmanager, expected)

    def test_accessing_passmanager_by_range_with_condition(self):
        """test accessing PassManager's passes by range with condition"""
        self.passmanager.append(PassB_TP_RA_PA())
        self.passmanager.append(PassE_AP_NR_NP(True))
        self.passmanager.append(
            ConditionalController(
                PassA_TP_NR_NP(),
                condition=lambda property_set: property_set["property"],
            )
        )
        self.passmanager.append(PassB_TP_RA_PA())

        new_passmanager = self.passmanager[1:3]

        expected = [
            "run analysis pass PassE_AP_NR_NP",
            "set property as True",
            "run transformation pass PassA_TP_NR_NP",
        ]
        self.assertScheduler(self.circuit, new_passmanager, expected)

    def test_accessing_passmanager_error(self):
        """testing accessing a pass item not in list"""
        self.passmanager.append(PassB_TP_RA_PA())

        with self.assertRaises(IndexError):
            self.passmanager = self.passmanager[99]


class TestPassManagerConcatenation(SchedulerTestCase):
    """test PassManager concatenation by + operator."""

    def setUp(self):
        super().setUp()
        self.passmanager1 = PassManager()
        self.passmanager2 = PassManager()
        self.circuit = QuantumCircuit(QuantumRegister(1))

    def test_concatenating_passmanagers(self):
        """test adding two PassManagers together"""
        self.passmanager1.append(PassB_TP_RA_PA())
        self.passmanager2.append(PassC_TP_RA_PA())

        new_passmanager = self.passmanager1 + self.passmanager2

        expected = [
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassB_TP_RA_PA",
            "run transformation pass PassC_TP_RA_PA",
        ]
        self.assertScheduler(self.circuit, new_passmanager, expected)

    def test_concatenating_passmanagers_with_condition(self):
        """test adding two pass managers with condition"""
        self.passmanager1.append(PassE_AP_NR_NP(True))
        self.passmanager1.append(PassB_TP_RA_PA())
        self.passmanager2.append(
            ConditionalController(
                PassC_TP_RA_PA(), condition=lambda property_set: property_set["property"]
            )
        )
        self.passmanager2.append(PassB_TP_RA_PA())

        new_passmanager = self.passmanager1 + self.passmanager2

        expected = [
            "run analysis pass PassE_AP_NR_NP",
            "set property as True",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassB_TP_RA_PA",
            "run transformation pass PassC_TP_RA_PA",
            "run transformation pass PassB_TP_RA_PA",
        ]
        self.assertScheduler(self.circuit, new_passmanager, expected)

    def test_adding_pass_to_passmanager(self):
        """test adding a pass to PassManager"""
        self.passmanager1.append(PassE_AP_NR_NP(argument1=1))
        self.passmanager1.append(PassB_TP_RA_PA())

        self.passmanager1 += PassC_TP_RA_PA()

        expected = [
            "run analysis pass PassE_AP_NR_NP",
            "set property as 1",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassB_TP_RA_PA",
            "run transformation pass PassC_TP_RA_PA",
        ]
        self.assertScheduler(self.circuit, self.passmanager1, expected)

    def test_adding_list_of_passes_to_passmanager(self):
        """test adding a list of passes to PassManager"""
        self.passmanager1.append(PassE_AP_NR_NP(argument1=1))
        self.passmanager1.append(PassB_TP_RA_PA())

        self.passmanager1 += [PassC_TP_RA_PA(), PassB_TP_RA_PA()]

        expected = [
            "run analysis pass PassE_AP_NR_NP",
            "set property as 1",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassB_TP_RA_PA",
            "run transformation pass PassC_TP_RA_PA",
            "run transformation pass PassB_TP_RA_PA",
        ]
        self.assertScheduler(self.circuit, self.passmanager1, expected)

    def test_adding_list_of_passes_to_passmanager_with_condition(self):
        """test adding a list of passes to a PassManager that have conditions"""
        self.passmanager1.append(PassE_AP_NR_NP(False))
        self.passmanager1.append(
            ConditionalController(
                PassB_TP_RA_PA(), condition=lambda property_set: property_set["property"]
            )
        )

        self.passmanager1 += PassC_TP_RA_PA()

        expected = [
            "run analysis pass PassE_AP_NR_NP",
            "set property as False",
            "run transformation pass PassA_TP_NR_NP",
            "run transformation pass PassC_TP_RA_PA",
        ]
        self.assertScheduler(self.circuit, self.passmanager1, expected)

    def test_adding_pass_to_passmanager_error(self):
        """testing adding a non-pass item to PassManager"""
        with self.assertRaises(TypeError):
            self.passmanager1 += "not a pass"

    def test_adding_list_to_passmanager_error(self):
        """testing adding a list having a non-pass item to PassManager"""
        with self.assertRaises(TypeError):
            self.passmanager1 += [PassB_TP_RA_PA(), "not a pass"]


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test the legacy Scheduling passes"""

import unittest

from ddt import ddt, data, unpack

from qiskit import QuantumCircuit
from qiskit.circuit import Delay, Parameter
from qiskit.circuit.library.standard_gates import XGate, YGate, CXGate
from qiskit.transpiler.exceptions import TranspilerError
from qiskit.transpiler.instruction_durations import InstructionDurations
from qiskit.transpiler.passes import ASAPSchedule, ALAPSchedule, DynamicalDecoupling
from qiskit.transpiler.passmanager import PassManager
from qiskit.transpiler.target import Target, InstructionProperties
from test import QiskitTestCase  # pylint: disable=wrong-import-order


@ddt
class TestSchedulingPass(QiskitTestCase):
    """Tests the Scheduling passes"""

    def test_alap_agree_with_reverse_asap_reverse(self):
        """Test if ALAP schedule agrees with doubly-reversed ASAP schedule."""
        qc = QuantumCircuit(2)
        qc.h(0)
        qc.delay(500, 1)
        qc.cx(0, 1)
        qc.measure_all()

        durations = InstructionDurations(
            [("h", 0, 200), ("cx", [0, 1], 700), ("measure", None, 1000)]
        )

        with self.assertWarns(DeprecationWarning):
            pm = PassManager(ALAPSchedule(durations))
        alap_qc = pm.run(qc)

        with self.assertWarns(DeprecationWarning):
            pm = PassManager(ASAPSchedule(durations))
        new_qc = pm.run(qc.reverse_ops())
        new_qc = new_qc.reverse_ops()
        new_qc.name = new_qc.name

        self.assertEqual(alap_qc, new_qc)

    @data(ALAPSchedule, ASAPSchedule)
    def test_classically_controlled_gate_after_measure(self, schedule_pass):
        """Test if ALAP/ASAP schedules circuits with c_if after measure with a common clbit.
        See: https://github.com/Qiskit/qiskit-terra/issues/7654

        (input)
             ┌─┐
        q_0: ┤M├───────────
             └╥┘   ┌───┐
        q_1: ─╫────┤ X ├───
              ║    └─╥─┘
              ║ ┌────╨────┐
        c: 1/═╩═╡ c_0 = T ╞
              0 └─────────┘

        (scheduled)
                                ┌─┐┌────────────────┐
        q_0: ───────────────────┤M├┤ Delay(200[dt]) ├
             ┌─────────────────┐└╥┘└─────┬───┬──────┘
        q_1: ┤ Delay(1000[dt]) ├─╫───────┤ X ├───────
             └─────────────────┘ ║       └─╥─┘
                                 ║    ┌────╨────┐
        c: 1/════════════════════╩════╡ c_0=0x1 ╞════
                                 0    └─────────┘
        """
        qc = QuantumCircuit(2, 1)
        qc.measure(0, 0)
        with self.assertWarns(DeprecationWarning):
            qc.x(1).c_if(0, True)

        durations = InstructionDurations([("x", None, 200), ("measure", None, 1000)])
        with self.assertWarns(DeprecationWarning):
            pm = PassManager(schedule_pass(durations))
        scheduled = pm.run(qc)

        expected = QuantumCircuit(2, 1)
        expected.measure(0, 0)
        expected.delay(1000, 1)  # x.c_if starts after measure
        with self.assertWarns(DeprecationWarning):
            expected.x(1).c_if(0, True)
        expected.delay(200, 0)

        self.assertEqual(expected, scheduled)

    @data(ALAPSchedule, ASAPSchedule)
    def test_measure_after_measure(self, schedule_pass):
        """Test if ALAP/ASAP schedules circuits with measure after measure with a common clbit.
        See: https://github.com/Qiskit/qiskit-terra/issues/7654

        (input)
             ┌───┐┌─┐
        q_0: ┤ X ├┤M├───
             └───┘└╥┘┌─┐
        q_1: ──────╫─┤M├
                   ║ └╥┘
        c: 1/══════╩══╩═
                   0  0

        (scheduled)
                    ┌───┐       ┌─┐┌─────────────────┐
        q_0: ───────┤ X ├───────┤M├┤ Delay(1000[dt]) ├
             ┌──────┴───┴──────┐└╥┘└───────┬─┬───────┘
        q_1: ┤ Delay(1200[dt]) ├─╫─────────┤M├────────
             └─────────────────┘ ║         └╥┘
        c: 1/════════════════════╩══════════╩═════════
                                 0          0
        """
        qc = QuantumCircuit(2, 1)
        qc.x(0)
        qc.measure(0, 0)
        qc.measure(1, 0)

        durations = InstructionDurations([("x", None, 200), ("measure", None, 1000)])
        with self.assertWarns(DeprecationWarning):
            pm = PassManager(schedule_pass(durations))
        scheduled = pm.run(qc)

        expected = QuantumCircuit(2, 1)
        expected.x(0)
        expected.measure(0, 0)
        expected.delay(1200, 1)
        expected.measure(1, 0)
        expected.delay(1000, 0)

        self.assertEqual(expected, scheduled)

    @data(ALAPSchedule, ASAPSchedule)
    def test_c_if_on_different_qubits(self, schedule_pass):
        """Test if ALAP/ASAP schedules circuits with `c_if`s on different qubits.

        (input)
             ┌─┐
        q_0: ┤M├──────────────────────
             └╥┘   ┌───┐
        q_1: ─╫────┤ X ├──────────────
              ║    └─╥─┘      ┌───┐
        q_2: ─╫──────╫────────┤ X ├───
              ║      ║        └─╥─┘
              ║ ┌────╨────┐┌────╨────┐
        c: 1/═╩═╡ c_0 = T ╞╡ c_0 = T ╞
              0 └─────────┘└─────────┘

        (scheduled)

                                ┌─┐┌────────────────┐
        q_0: ───────────────────┤M├┤ Delay(200[dt]) ├───────────
             ┌─────────────────┐└╥┘└─────┬───┬──────┘
        q_1: ┤ Delay(1000[dt]) ├─╫───────┤ X ├──────────────────
             ├─────────────────┤ ║       └─╥─┘          ┌───┐
        q_2: ┤ Delay(1000[dt]) ├─╫─────────╫────────────┤ X ├───
             └─────────────────┘ ║         ║            └─╥─┘
                                 ║    ┌────╨────┐    ┌────╨────┐
        c: 1/════════════════════╩════╡ c_0=0x1 ╞════╡ c_0=0x1 ╞
                                 0    └─────────┘    └─────────┘
        """
        qc = QuantumCircuit(3, 1)
        qc.measure(0, 0)
        with self.assertWarns(DeprecationWarning):
            qc.x(1).c_if(0, True)
        with self.assertWarns(DeprecationWarning):
            qc.x(2).c_if(0, True)

        durations = InstructionDurations([("x", None, 200), ("measure", None, 1000)])
        with self.assertWarns(DeprecationWarning):
            pm = PassManager(schedule_pass(durations))
        scheduled = pm.run(qc)

        expected = QuantumCircuit(3, 1)
        expected.measure(0, 0)
        expected.delay(1000, 1)
        expected.delay(1000, 2)
        with self.assertWarns(DeprecationWarning):
            expected.x(1).c_if(0, True)
        with self.assertWarns(DeprecationWarning):
            expected.x(2).c_if(0, True)
        expected.delay(200, 0)

        self.assertEqual(expected, scheduled)

    @data(ALAPSchedule, ASAPSchedule)
    def test_shorter_measure_after_measure(self, schedule_pass):
        """Test if ALAP/ASAP schedules circuits with shorter measure after measure with a common clbit.

        (input)
             ┌─┐
        q_0: ┤M├───
             └╥┘┌─┐
        q_1: ─╫─┤M├
              ║ └╥┘
        c: 1/═╩══╩═
              0  0

        (scheduled)
                                ┌─┐┌────────────────┐
        q_0: ───────────────────┤M├┤ Delay(700[dt]) ├
             ┌─────────────────┐└╥┘└──────┬─┬───────┘
        q_1: ┤ Delay(1000[dt]) ├─╫────────┤M├────────
             └─────────────────┘ ║        └╥┘
        c: 1/════════════════════╩═════════╩═════════
                                 0         0
        """
        qc = QuantumCircuit(2, 1)
        qc.measure(0, 0)
        qc.measure(1, 0)

        durations = InstructionDurations([("measure", [0], 1000), ("measure", [1], 700)])
        with self.assertWarns(DeprecationWarning):
            pm = PassManager(schedule_pass(durations))
        scheduled = pm.run(qc)

        expected = QuantumCircuit(2, 1)
        expected.measure(0, 0)
        expected.delay(1000, 1)
        expected.measure(1, 0)
        expected.delay(700, 0)

        self.assertEqual(expected, scheduled)

    @data(ALAPSchedule, ASAPSchedule)
    def test_measure_after_c_if(self, schedule_pass):
        """Test if ALAP/ASAP schedules circuits with c_if after measure with a common clbit.

        (input)
             ┌─┐
        q_0: ┤M├──────────────
             └╥┘   ┌───┐
        q_1: ─╫────┤ X ├──────
              ║    └─╥─┘   ┌─┐
        q_2: ─╫──────╫─────┤M├
              ║ ┌────╨────┐└╥┘
        c: 1/═╩═╡ c_0 = T ╞═╩═
              0 └─────────┘ 0

        (scheduled)
                                ┌─┐┌─────────────────┐
        q_0: ───────────────────┤M├┤ Delay(1000[dt]) ├──────────────────
             ┌─────────────────┐└╥┘└──────┬───┬──────┘┌────────────────┐
        q_1: ┤ Delay(1000[dt]) ├─╫────────┤ X ├───────┤ Delay(800[dt]) ├
             ├─────────────────┤ ║        └─╥─┘       └──────┬─┬───────┘
        q_2: ┤ Delay(1000[dt]) ├─╫──────────╫────────────────┤M├────────
             └─────────────────┘ ║     ┌────╨────┐           └╥┘
        c: 1/════════════════════╩═════╡ c_0=0x1 ╞════════════╩═════════
                                 0     └─────────┘            0
        """
        qc = QuantumCircuit(3, 1)
        qc.measure(0, 0)
        with self.assertWarns(DeprecationWarning):
            qc.x(1).c_if(0, 1)
        qc.measure(2, 0)

        durations = InstructionDurations([("x", None, 200), ("measure", None, 1000)])
        with self.assertWarns(DeprecationWarning):
            pm = PassManager(schedule_pass(durations))
        scheduled = pm.run(qc)

        expected = QuantumCircuit(3, 1)
        expected.delay(1000, 1)
        expected.delay(1000, 2)
        expected.measure(0, 0)
        with self.assertWarns(DeprecationWarning):
            expected.x(1).c_if(0, 1)
        expected.measure(2, 0)
        expected.delay(1000, 0)
        expected.delay(800, 1)

        self.assertEqual(expected, scheduled)

    def test_parallel_gate_different_length(self):
        """Test circuit having two parallel instruction with different length.

        (input)
             ┌───┐┌─┐
        q_0: ┤ X ├┤M├───
             ├───┤└╥┘┌─┐
        q_1: ┤ X ├─╫─┤M├
             └───┘ ║ └╥┘
        c: 2/══════╩══╩═
                   0  1

        (expected, ALAP)
             ┌────────────────┐┌───┐┌─┐
        q_0: ┤ Delay(200[dt]) ├┤ X ├┤M├
             └─────┬───┬──────┘└┬─┬┘└╥┘
        q_1: ──────┤ X ├────────┤M├──╫─
                   └───┘        └╥┘  ║
        c: 2/════════════════════╩═══╩═
                                 1   0

        (expected, ASAP)
             ┌───┐┌─┐┌────────────────┐
        q_0: ┤ X ├┤M├┤ Delay(200[dt]) ├
             ├───┤└╥┘└──────┬─┬───────┘
        q_1: ┤ X ├─╫────────┤M├────────
             └───┘ ║        └╥┘
        c: 2/══════╩═════════╩═════════
                   0         1

        """
        qc = QuantumCircuit(2, 2)
        qc.x(0)
        qc.x(1)
        qc.measure(0, 0)
        qc.measure(1, 1)

        durations = InstructionDurations(
            [("x", [0], 200), ("x", [1], 400), ("measure", None, 1000)]
        )
        with self.assertWarns(DeprecationWarning):
            pm = PassManager(ALAPSchedule(durations))
        qc_alap = pm.run(qc)

        alap_expected = QuantumCircuit(2, 2)
        alap_expected.delay(200, 0)
        alap_expected.x(0)
        alap_expected.x(1)
        alap_expected.measure(0, 0)
        alap_expected.measure(1, 1)

        self.assertEqual(qc_alap, alap_expected)

        with self.assertWarns(DeprecationWarning):
            pm = PassManager(ASAPSchedule(durations))
        qc_asap = pm.run(qc)

        asap_expected = QuantumCircuit(2, 2)
        asap_expected.x(0)
        asap_expected.x(1)
        asap_expected.measure(0, 0)  # immediately start after X gate
        asap_expected.measure(1, 1)
        asap_expected.delay(200, 0)

        self.assertEqual(qc_asap, asap_expected)

    def test_parallel_gate_different_length_with_barrier(self):
        """Test circuit having two parallel instruction with different length with barrier.

        (input)
             ┌───┐┌─┐
        q_0: ┤ X ├┤M├───
             ├───┤└╥┘┌─┐
        q_1: ┤ X ├─╫─┤M├
             └───┘ ║ └╥┘
        c: 2/══════╩══╩═
                   0  1

        (expected, ALAP)
             ┌────────────────┐┌───┐ ░ ┌─┐
        q_0: ┤ Delay(200[dt]) ├┤ X ├─░─┤M├───
             └─────┬───┬──────┘└───┘ ░ └╥┘┌─┐
        q_1: ──────┤ X ├─────────────░──╫─┤M├
                   └───┘             ░  ║ └╥┘
        c: 2/═══════════════════════════╩══╩═
                                        0  1

        (expected, ASAP)
             ┌───┐┌────────────────┐ ░ ┌─┐
        q_0: ┤ X ├┤ Delay(200[dt]) ├─░─┤M├───
             ├───┤└────────────────┘ ░ └╥┘┌─┐
        q_1: ┤ X ├───────────────────░──╫─┤M├
             └───┘                   ░  ║ └╥┘
        c: 2/═══════════════════════════╩══╩═
                                        0  1
        """
        qc = QuantumCircuit(2, 2)
        qc.x(0)
        qc.x(1)
        qc.barrier()
        qc.measure(0, 0)
        qc.measure(1, 1)

        durations = InstructionDurations(
            [("x", [0], 200), ("x", [1], 400), ("measure", None, 1000)]
        )
        with self.assertWarns(DeprecationWarning):
            pm = PassManager(ALAPSchedule(durations))
        qc_alap = pm.run(qc)

        alap_expected = QuantumCircuit(2, 2)
        alap_expected.delay(200, 0)
        alap_expected.x(0)
        alap_expected.x(1)
        alap_expected.barrier()
        alap_expected.measure(0, 0)
        alap_expected.measure(1, 1)

        self.assertEqual(qc_alap, alap_expected)

        with self.assertWarns(DeprecationWarning):
            pm = PassManager(ASAPSchedule(durations))
        qc_asap = pm.run(qc)

        asap_expected = QuantumCircuit(2, 2)
        asap_expected.x(0)
        asap_expected.delay(200, 0)
        asap_expected.x(1)
        asap_expected.barrier()
        asap_expected.measure(0, 0)
        asap_expected.measure(1, 1)

        self.assertEqual(qc_asap, asap_expected)

    def test_measure_after_c_if_on_edge_locking(self):
        """Test if ALAP/ASAP schedules circuits with c_if after measure with a common clbit.

        The scheduler is configured to reproduce behavior of the 0.20.0,
        in which clbit lock is applied to the end-edge of measure instruction.
        See https://github.com/Qiskit/qiskit-terra/pull/7655

        (input)
             ┌─┐
        q_0: ┤M├──────────────
             └╥┘   ┌───┐
        q_1: ─╫────┤ X ├──────
              ║    └─╥─┘   ┌─┐
        q_2: ─╫──────╫─────┤M├
              ║ ┌────╨────┐└╥┘
        c: 1/═╩═╡ c_0 = T ╞═╩═
              0 └─────────┘ 0

        (ASAP scheduled)
                                ┌─┐┌────────────────┐
        q_0: ───────────────────┤M├┤ Delay(200[dt]) ├─────────────────────
             ┌─────────────────┐└╥┘└─────┬───┬──────┘
        q_1: ┤ Delay(1000[dt]) ├─╫───────┤ X ├────────────────────────────
             └─────────────────┘ ║       └─╥─┘       ┌─┐┌────────────────┐
        q_2: ────────────────────╫─────────╫─────────┤M├┤ Delay(200[dt]) ├
                                 ║    ┌────╨────┐    └╥┘└────────────────┘
        c: 1/════════════════════╩════╡ c_0=0x1 ╞═════╩═══════════════════
                                 0    └─────────┘     0

        (ALAP scheduled)
                                ┌─┐┌────────────────┐
        q_0: ───────────────────┤M├┤ Delay(200[dt]) ├───
             ┌─────────────────┐└╥┘└─────┬───┬──────┘
        q_1: ┤ Delay(1000[dt]) ├─╫───────┤ X ├──────────
             └┬────────────────┤ ║       └─╥─┘       ┌─┐
        q_2: ─┤ Delay(200[dt]) ├─╫─────────╫─────────┤M├
              └────────────────┘ ║    ┌────╨────┐    └╥┘
        c: 1/════════════════════╩════╡ c_0=0x1 ╞═════╩═
                                 0    └─────────┘     0

        """
        qc = QuantumCircuit(3, 1)
        qc.measure(0, 0)
        with self.assertWarns(DeprecationWarning):
            qc.x(1).c_if(0, 1)
        qc.measure(2, 0)

        durations = InstructionDurations([("x", None, 200), ("measure", None, 1000)])

        # lock at the end edge
        with self.assertWarns(DeprecationWarning):
            actual_asap = PassManager(ASAPSchedule(durations, clbit_write_latency=1000)).run(qc)
            actual_alap = PassManager(ALAPSchedule(durations, clbit_write_latency=1000)).run(qc)

        # start times of 2nd measure depends on ASAP/ALAP
        expected_asap = QuantumCircuit(3, 1)
        expected_asap.measure(0, 0)
        expected_asap.delay(1000, 1)
        with self.assertWarns(DeprecationWarning):
            expected_asap.x(1).c_if(0, 1)
        expected_asap.measure(2, 0)
        expected_asap.delay(200, 0)
        expected_asap.delay(200, 2)
        self.assertEqual(expected_asap, actual_asap)

        expected_alap = QuantumCircuit(3, 1)
        expected_alap.measure(0, 0)
        expected_alap.delay(1000, 1)
        with self.assertWarns(DeprecationWarning):
            expected_alap.x(1).c_if(0, 1)
        expected_alap.delay(200, 2)
        expected_alap.measure(2, 0)
        expected_alap.delay(200, 0)
        self.assertEqual(expected_alap, actual_alap)

    @data([100, 200], [500, 0], [1000, 200])
    @unpack
    def test_active_reset_circuit(self, write_lat, cond_lat):
        """Test practical example of reset circuit.

        Because of the stimulus pulse overlap with the previous XGate on the q register,
        measure instruction is always triggered after XGate regardless of write latency.
        Thus only conditional latency matters in the scheduling.

        (input)
             ┌─┐   ┌───┐   ┌─┐   ┌───┐   ┌─┐   ┌───┐
          q: ┤M├───┤ X ├───┤M├───┤ X ├───┤M├───┤ X ├───
             └╥┘   └─╥─┘   └╥┘   └─╥─┘   └╥┘   └─╥─┘
              ║ ┌────╨────┐ ║ ┌────╨────┐ ║ ┌────╨────┐
        c: 1/═╩═╡ c_0=0x1 ╞═╩═╡ c_0=0x1 ╞═╩═╡ c_0=0x1 ╞
              0 └─────────┘ 0 └─────────┘ 0 └─────────┘

        """
        qc = QuantumCircuit(1, 1)
        qc.measure(0, 0)
        with self.assertWarns(DeprecationWarning):
            qc.x(0).c_if(0, 1)
        qc.measure(0, 0)
        with self.assertWarns(DeprecationWarning):
            qc.x(0).c_if(0, 1)
        qc.measure(0, 0)
        with self.assertWarns(DeprecationWarning):
            qc.x(0).c_if(0, 1)

        durations = InstructionDurations([("x", None, 100), ("measure", None, 1000)])
        with self.assertWarns(DeprecationWarning):
            actual_asap = PassManager(
                ASAPSchedule(durations, clbit_write_latency=write_lat, conditional_latency=cond_lat)
            ).run(qc)
            actual_alap = PassManager(
                ALAPSchedule(durations, clbit_write_latency=write_lat, conditional_latency=cond_lat)
            ).run(qc)

        expected = QuantumCircuit(1, 1)
        expected.measure(0, 0)
        if cond_lat > 0:
            expected.delay(cond_lat, 0)
        with self.assertWarns(DeprecationWarning):
            expected.x(0).c_if(0, 1)
        expected.measure(0, 0)
        if cond_lat > 0:
            expected.delay(cond_lat, 0)
        with self.assertWarns(DeprecationWarning):
            expected.x(0).c_if(0, 1)
        expected.measure(0, 0)
        if cond_lat > 0:
            expected.delay(cond_lat, 0)
        with self.assertWarns(DeprecationWarning):
            expected.x(0).c_if(0, 1)

        self.assertEqual(expected, actual_asap)
        self.assertEqual(expected, actual_alap)

    def test_random_complicated_circuit(self):
        """Test scheduling complicated circuit with control flow.

        (input)
             ┌────────────────┐   ┌───┐    ░                  ┌───┐   »
        q_0: ┤ Delay(100[dt]) ├───┤ X ├────░──────────────────┤ X ├───»
             └────────────────┘   └─╥─┘    ░       ┌───┐      └─╥─┘   »
        q_1: ───────────────────────╫──────░───────┤ X ├────────╫─────»
                                    ║      ░ ┌─┐   └─╥─┘        ║     »
        q_2: ───────────────────────╫──────░─┤M├─────╫──────────╫─────»
                               ┌────╨────┐ ░ └╥┘┌────╨────┐┌────╨────┐»
        c: 1/══════════════════╡ c_0=0x1 ╞════╩═╡ c_0=0x0 ╞╡ c_0=0x0 ╞»
                               └─────────┘    0 └─────────┘└─────────┘»
        «     ┌────────────────┐┌───┐
        «q_0: ┤ Delay(300[dt]) ├┤ X ├─────■─────
        «     └────────────────┘└───┘   ┌─┴─┐
        «q_1: ────────■─────────────────┤ X ├───
        «           ┌─┴─┐        ┌─┐    └─╥─┘
        «q_2: ──────┤ X ├────────┤M├──────╫─────
        «           └───┘        └╥┘ ┌────╨────┐
        «c: 1/════════════════════╩══╡ c_0=0x0 ╞
        «                         0  └─────────┘

        (ASAP scheduled) duration = 2800 dt
             ┌────────────────┐┌────────────────┐   ┌───┐    ░ ┌─────────────────┐»
        q_0: ┤ Delay(100[dt]) ├┤ Delay(100[dt]) ├───┤ X ├────░─┤ Delay(1400[dt]) ├»
             ├────────────────┤└────────────────┘   └─╥─┘    ░ ├─────────────────┤»
        q_1: ┤ Delay(300[dt]) ├───────────────────────╫──────░─┤ Delay(1200[dt]) ├»
             ├────────────────┤                       ║      ░ └───────┬─┬───────┘»
        q_2: ┤ Delay(300[dt]) ├───────────────────────╫──────░─────────┤M├────────»
             └────────────────┘                  ┌────╨────┐ ░         └╥┘        »
        c: 1/════════════════════════════════════╡ c_0=0x1 ╞════════════╩═════════»
                                                 └─────────┘            0         »
        «                                     ┌───┐   ┌────────────────┐»
        «q_0: ────────────────────────────────┤ X ├───┤ Delay(300[dt]) ├»
        «        ┌───┐                        └─╥─┘   └────────────────┘»
        «q_1: ───┤ X ├──────────────────────────╫─────────────■─────────»
        «        └─╥─┘   ┌────────────────┐     ║           ┌─┴─┐       »
        «q_2: ─────╫─────┤ Delay(300[dt]) ├─────╫───────────┤ X ├───────»
        «     ┌────╨────┐└────────────────┘┌────╨────┐      └───┘       »
        «c: 1/╡ c_0=0x0 ╞══════════════════╡ c_0=0x0 ╞══════════════════»
        «     └─────────┘                  └─────────┘                  »
        «           ┌───┐                  ┌────────────────┐
        «q_0: ──────┤ X ├────────────■─────┤ Delay(700[dt]) ├
        «     ┌─────┴───┴──────┐   ┌─┴─┐   ├────────────────┤
        «q_1: ┤ Delay(400[dt]) ├───┤ X ├───┤ Delay(700[dt]) ├
        «     ├────────────────┤   └─╥─┘   └──────┬─┬───────┘
        «q_2: ┤ Delay(300[dt]) ├─────╫────────────┤M├────────
        «     └────────────────┘┌────╨────┐       └╥┘
        «c: 1/══════════════════╡ c_0=0x0 ╞════════╩═════════
        «                       └─────────┘        0

        (ALAP scheduled) duration = 3100
             ┌────────────────┐┌────────────────┐   ┌───┐    ░ ┌─────────────────┐»
        q_0: ┤ Delay(100[dt]) ├┤ Delay(100[dt]) ├───┤ X ├────░─┤ Delay(1400[dt]) ├»
             ├────────────────┤└────────────────┘   └─╥─┘    ░ ├─────────────────┤»
        q_1: ┤ Delay(300[dt]) ├───────────────────────╫──────░─┤ Delay(1200[dt]) ├»
             ├────────────────┤                       ║      ░ └───────┬─┬───────┘»
        q_2: ┤ Delay(300[dt]) ├───────────────────────╫──────░─────────┤M├────────»
             └────────────────┘                  ┌────╨────┐ ░         └╥┘        »
        c: 1/════════════════════════════════════╡ c_0=0x1 ╞════════════╩═════════»
                                                 └─────────┘            0         »
        «                                     ┌───┐   ┌────────────────┐»
        «q_0: ────────────────────────────────┤ X ├───┤ Delay(300[dt]) ├»
        «        ┌───┐   ┌────────────────┐   └─╥─┘   └────────────────┘»
        «q_1: ───┤ X ├───┤ Delay(300[dt]) ├─────╫─────────────■─────────»
        «        └─╥─┘   ├────────────────┤     ║           ┌─┴─┐       »
        «q_2: ─────╫─────┤ Delay(600[dt]) ├─────╫───────────┤ X ├───────»
        «     ┌────╨────┐└────────────────┘┌────╨────┐      └───┘       »
        «c: 1/╡ c_0=0x0 ╞══════════════════╡ c_0=0x0 ╞══════════════════»
        «     └─────────┘                  └─────────┘                  »
        «           ┌───┐                  ┌────────────────┐
        «q_0: ──────┤ X ├────────────■─────┤ Delay(700[dt]) ├
        «     ┌─────┴───┴──────┐   ┌─┴─┐   ├────────────────┤
        «q_1: ┤ Delay(100[dt]) ├───┤ X ├───┤ Delay(700[dt]) ├
        «     └──────┬─┬───────┘   └─╥─┘   └────────────────┘
        «q_2: ───────┤M├─────────────╫───────────────────────
        «            └╥┘        ┌────╨────┐
        «c: 1/════════╩═════════╡ c_0=0x0 ╞══════════════════
        «             0         └─────────┘

        """
        qc = QuantumCircuit(3, 1)
        qc.delay(100, 0)
        with self.assertWarns(DeprecationWarning):
            qc.x(0).c_if(0, 1)
        qc.barrier()
        qc.measure(2, 0)
        with self.assertWarns(DeprecationWarning):
            qc.x(1).c_if(0, 0)
        with self.assertWarns(DeprecationWarning):
            qc.x(0).c_if(0, 0)
        qc.delay(300, 0)
        qc.cx(1, 2)
        qc.x(0)
        with self.assertWarns(DeprecationWarning):
            qc.cx(0, 1).c_if(0, 0)
        qc.measure(2, 0)

        durations = InstructionDurations(
            [("x", None, 100), ("measure", None, 1000), ("cx", None, 200)]
        )

        with self.assertWarns(DeprecationWarning):
            actual_asap = PassManager(
                ASAPSchedule(durations, clbit_write_latency=100, conditional_latency=200)
            ).run(qc)
            actual_alap = PassManager(
                ALAPSchedule(durations, clbit_write_latency=100, conditional_latency=200)
            ).run(qc)

        expected_asap = QuantumCircuit(3, 1)
        expected_asap.delay(100, 0)
        expected_asap.delay(100, 0)  # due to conditional latency of 200dt
        expected_asap.delay(300, 1)
        expected_asap.delay(300, 2)
        with self.assertWarns(DeprecationWarning):
            expected_asap.x(0).c_if(0, 1)
        expected_asap.barrier()
        expected_asap.delay(1400, 0)
        expected_asap.delay(1200, 1)
        expected_asap.measure(2, 0)
        with self.assertWarns(DeprecationWarning):
            expected_asap.x(1).c_if(0, 0)
        with self.assertWarns(DeprecationWarning):
            expected_asap.x(0).c_if(0, 0)
        expected_asap.delay(300, 0)
        expected_asap.x(0)
        expected_asap.delay(300, 2)
        expected_asap.cx(1, 2)
        expected_asap.delay(400, 1)
        with self.assertWarns(DeprecationWarning):
            expected_asap.cx(0, 1).c_if(0, 0)
        expected_asap.delay(700, 0)  # creg is released at t0 of cx(0,1).c_if(0,0)
        expected_asap.delay(
            700, 1
        )  # no creg write until 100dt. thus measure can move left by 300dt.
        expected_asap.delay(300, 2)
        expected_asap.measure(2, 0)
        self.assertEqual(expected_asap, actual_asap)
        self.assertEqual(actual_asap.duration, 3100)

        expected_alap = QuantumCircuit(3, 1)
        expected_alap.delay(100, 0)
        expected_alap.delay(100, 0)  # due to conditional latency of 200dt
        expected_alap.delay(300, 1)
        expected_alap.delay(300, 2)
        with self.assertWarns(DeprecationWarning):
            expected_alap.x(0).c_if(0, 1)
        expected_alap.barrier()
        expected_alap.delay(1400, 0)
        expected_alap.delay(1200, 1)
        expected_alap.measure(2, 0)
        with self.assertWarns(DeprecationWarning):
            expected_alap.x(1).c_if(0, 0)
        with self.assertWarns(DeprecationWarning):
            expected_alap.x(0).c_if(0, 0)
        expected_alap.delay(300, 0)
        expected_alap.x(0)
        expected_alap.delay(300, 1)
        expected_alap.delay(600, 2)
        expected_alap.cx(1, 2)
        expected_alap.delay(100, 1)
        with self.assertWarns(DeprecationWarning):
            expected_alap.cx(0, 1).c_if(0, 0)
        expected_alap.measure(2, 0)
        expected_alap.delay(700, 0)
        expected_alap.delay(700, 1)
        self.assertEqual(expected_alap, actual_alap)
        self.assertEqual(actual_alap.duration, 3100)

    def test_dag_introduces_extra_dependency_between_conditionals(self):
        """Test dependency between conditional operations in the scheduling.

        In the below example circuit, the conditional x on q1 could start at time 0,
        however it must be scheduled after the conditional x on q0 in ASAP scheduling.
        That is because circuit model used in the transpiler passes (DAGCircuit)
        interprets instructions acting on common clbits must be run in the order
        given by the original circuit (QuantumCircuit).

        (input)
             ┌────────────────┐   ┌───┐
        q_0: ┤ Delay(100[dt]) ├───┤ X ├───
             └─────┬───┬──────┘   └─╥─┘
        q_1: ──────┤ X ├────────────╫─────
                   └─╥─┘            ║
                ┌────╨────┐    ┌────╨────┐
        c: 1/═══╡ c_0=0x1 ╞════╡ c_0=0x1 ╞
                └─────────┘    └─────────┘

        (ASAP scheduled)
             ┌────────────────┐   ┌───┐
        q_0: ┤ Delay(100[dt]) ├───┤ X ├──────────────
             ├────────────────┤   └─╥─┘      ┌───┐
        q_1: ┤ Delay(100[dt]) ├─────╫────────┤ X ├───
             └────────────────┘     ║        └─╥─┘
                               ┌────╨────┐┌────╨────┐
        c: 1/══════════════════╡ c_0=0x1 ╞╡ c_0=0x1 ╞
                               └─────────┘└─────────┘
        """
        qc = QuantumCircuit(2, 1)
        qc.delay(100, 0)
        with self.assertWarns(DeprecationWarning):
            qc.x(0).c_if(0, True)
        with self.assertWarns(DeprecationWarning):
            qc.x(1).c_if(0, True)

        durations = InstructionDurations([("x", None, 160)])
        with self.assertWarns(DeprecationWarning):
            pm = PassManager(ASAPSchedule(durations))
        scheduled = pm.run(qc)

        expected = QuantumCircuit(2, 1)
        expected.delay(100, 0)
        expected.delay(100, 1)  # due to extra dependency on clbits
        with self.assertWarns(DeprecationWarning):
            expected.x(0).c_if(0, True)
        with self.assertWarns(DeprecationWarning):
            expected.x(1).c_if(0, True)

        self.assertEqual(expected, scheduled)

    @data(ALAPSchedule, ASAPSchedule)
    def test_respect_target_instruction_constraints(self, schedule_pass):
        """Test if ALAP/ASAP does not pad delays for qubits that do not support delay instructions.
        See: https://github.com/Qiskit/qiskit-terra/issues/9993
        """
        target = Target(dt=1)
        target.add_instruction(XGate(), {(1,): InstructionProperties(duration=200)})
        # delays are not supported

        qc = QuantumCircuit(2)
        qc.x(1)

        with self.assertWarns(DeprecationWarning):
            pm = PassManager(schedule_pass(target=target))
        scheduled = pm.run(qc)

        expected = QuantumCircuit(2)
        expected.x(1)
        # no delay on qubit 0

        self.assertEqual(expected, scheduled)

    def test_dd_respect_target_instruction_constraints(self):
        """Test if DD pass does not pad delays for qubits that do not support delay instructions
        and does not insert DD gates for qubits that do not support necessary gates.
        See: https://github.com/Qiskit/qiskit-terra/issues/9993
        """
        qc = QuantumCircuit(3)
        qc.cx(0, 1)
        qc.cx(1, 2)

        target = Target(dt=1)
        # Y is partially supported (not supported on qubit 2)
        target.add_instruction(
            XGate(), {(q,): InstructionProperties(duration=100) for q in range(2)}
        )
        target.add_instruction(
            CXGate(),
            {
                (0, 1): InstructionProperties(duration=1000),
                (1, 2): InstructionProperties(duration=1000),
            },
        )
        # delays are not supported

        # No DD instructions nor delays are padded due to no delay support in the target
        with self.assertWarns(DeprecationWarning):
            pm_scheduler = PassManager(
                [
                    ALAPSchedule(target=target),
                    DynamicalDecoupling(
                        durations=None, dd_sequence=[XGate(), XGate()], target=target
                    ),
                ]
            )
        scheduled = pm_scheduler.run(qc)
        self.assertEqual(qc, scheduled)

        # Fails since Y is not supported in the target
        with self.assertWarns(DeprecationWarning):
            with self.assertRaises(TranspilerError):
                PassManager(
                    [
                        ALAPSchedule(target=target),
                        DynamicalDecoupling(
                            durations=None,
                            dd_sequence=[XGate(), YGate(), XGate(), YGate()],
                            target=target,
                        ),
                    ]
                )

        # Add delay support to the target
        target.add_instruction(Delay(Parameter("t")), {(q,): None for q in range(3)})
        # No error but no DD on qubit 2 (just delay is padded) since X is not supported on it
        scheduled = pm_scheduler.run(qc)

        expected = QuantumCircuit(3)
        expected.delay(1000, [2])
        expected.cx(0, 1)
        expected.cx(1, 2)
        expected.delay(200, [0])
        expected.x([0])
        expected.delay(400, [0])
        expected.x([0])
        expected.delay(200, [0])
        self.assertEqual(expected, scheduled)


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
================================
OpenQASM 3 (:mod:`qiskit.qasm3`)
================================

.. currentmodule:: qiskit.qasm3

Qiskit provides some tools for converting between `OpenQASM 3 <https://openqasm.com>`__
representations of quantum programs, and the :class:`.QuantumCircuit` class.  These will continue to
evolve as Qiskit's support for the dynamic-circuit capabilities expressed by OpenQASM 3 increases.


Exporting to OpenQASM 3
=======================

The high-level functions are simply :func:`dump` and :func:`dumps`, which respectively export to a
file (given as a filename) and to a Python string.

.. autofunction:: dump
.. autofunction:: dumps

Both of these exporter functions are single-use wrappers around the main :class:`Exporter` class.
For more complex exporting needs, including dumping multiple circuits in a single session, it may be
more convenient or faster to use the complete interface.

.. autoclass:: Exporter
    :members:

All of these interfaces will raise :exc:`QASM3ExporterError` on failure.

.. autoexception:: QASM3ExporterError

Experimental features
---------------------

The OpenQASM 3 language is still evolving as hardware capabilities improve, so there is no final
syntax that Qiskit can reliably target.  In order to represent the evolving language, we will
sometimes release features before formal standardization, which may need to change as the review
process in the OpenQASM 3 design committees progresses.  By default, the exporters will only support
standardised features of the language.  To enable these early-release features, use the
``experimental`` keyword argument of :func:`dump` and :func:`dumps`.  The available feature flags
are:

.. autoclass:: ExperimentalFeatures
    :members:

If you want to enable multiple experimental features, you should combine the flags using the ``|``
operator, such as ``flag1 | flag2``.

For example, to perform an export using the early semantics of ``switch`` support::

    from qiskit import qasm3, QuantumCircuit, QuantumRegister, ClassicalRegister

    # Build the circuit
    qreg = QuantumRegister(3)
    creg = ClassicalRegister(3)
    qc = QuantumCircuit(qreg, creg)
    with qc.switch(creg) as case:
        with case(0):
            qc.x(0)
        with case(1, 2):
            qc.x(1)
        with case(case.DEFAULT):
            qc.x(2)

    # Export to an OpenQASM 3 string.
    qasm_string = qasm3.dumps(qc, experimental=qasm3.ExperimentalFeatures.SWITCH_CASE_V1)


.. note::

    All features enabled by the experimental flags are naturally transient.  If it becomes necessary
    to remove flags, they will be subject to `the standard Qiskit deprecation policy
    <https://github.com/Qiskit/qiskit/blob/main/DEPRECATION.md>`__.  We will leave these experimental
    flags in place for as long as is reasonable.

    However, we cannot guarantee any support windows for *consumers* of OpenQASM 3 code generated
    using these experimental flags, if the OpenQASM 3 language specification changes the proposal
    that the flag is based on.  It is possible that any tool you are using to consume OpenQASM 3
    code created using these flags may update or remove their support while Qiskit continues to
    offer the flag.  You should not rely on the resultant experimental OpenQASM 3 code for long-term
    storage of programs.


Importing from OpenQASM 3
=========================

Currently only two high-level functions are offered, as Qiskit support for importing from OpenQASM 3
is in its infancy, and the implementation is expected to change significantly.  The two functions
are :func:`load` and :func:`loads`, which are direct counterparts of :func:`dump` and :func:`dumps`,
respectively loading a program indirectly from a named file and directly from a given string.

.. note::

    While we are still in the exploratory release period, to use either function, the package
    ``qiskit_qasm3_import`` must be installed.  This can be done by installing Qiskit with the
    ``qasm3-import`` extra, such as by:

    .. code-block:: text

        pip install qiskit[qasm3-import]

    We expect that this functionality will eventually be merged into Qiskit, and no longer
    require an optional import, but we do not yet have a timeline for this.

.. autofunction:: load
.. autofunction:: loads

Both of these two functions raise :exc:`QASM3ImporterError` on failure.

.. autoexception:: QASM3ImporterError

For example, we can define a quantum program using OpenQASM 3, and use :func:`loads` to directly
convert it into a :class:`.QuantumCircuit`:

.. plot::
    :alt: Circuit diagram output by the previous code.
    :include-source:

    import qiskit.qasm3

    program = \"\"\"
        OPENQASM 3.0;
        include "stdgates.inc";

        input float[64] a;
        qubit[3] q;
        bit[2] mid;
        bit[3] out;

        let aliased = q[0:1];

        gate my_gate(a) c, t {
          gphase(a / 2);
          ry(a) c;
          cx c, t;
        }
        gate my_phase(a) c {
          ctrl @ inv @ gphase(a) c;
        }

        my_gate(a * 2) aliased[0], q[{1, 2}][0];
        measure q[0] -> mid[0];
        measure q[1] -> mid[1];

        while (mid == "00") {
          reset q[0];
          reset q[1];
          my_gate(a) q[0], q[1];
          my_phase(a - pi/2) q[1];
          mid[0] = measure q[0];
          mid[1] = measure q[1];
        }

        if (mid[0]) {
          let inner_alias = q[{0, 1}];
          reset inner_alias;
        }

        out = measure q;
    \"\"\"
    circuit = qiskit.qasm3.loads(program)
    circuit.draw("mpl")


Experimental import interface
-----------------------------

The import functions given above rely on the ANTLR-based reference parser from the OpenQASM project
itself, which is more intended as a language reference than a performant parser.  You need to have
the extension ``qiskit-qasm3-import`` installed to use it.

Qiskit is developing a native parser, written in Rust, which is available as part of the core Qiskit
package.  This parser is still in its early experimental stages, so is missing features and its
interface is changing and expanding, but it is typically orders of magnitude more performant for the
subset of OpenQASM 3 it currently supports, and its internals produce better error diagnostics on
parsing failures.

You can use the experimental interface immediately, with similar functions to the main interface
above:

.. autofunction:: load_experimental
.. autofunction:: loads_experimental

These two functions are both experimental, meaning they issue an :exc:`.ExperimentalWarning` on
usage, and their interfaces may be subject to change within the Qiskit 1.x release series.  In
particular, the native parser may be promoted to be the default version of :func:`load` and
:func:`loads`.  If you are happy to accept the risk of using the experimental interface, you can
disable the warning by doing::

    import warnings
    from qiskit.exceptions import ExperimentalWarning

    warnings.filterwarnings("ignore", category=ExperimentalWarning, module="qiskit.qasm3")

These two functions allow for specifying include paths as an iterable of paths, and for specifying
custom Python constructors to use for particular gates.  These custom constructors are specified by
using the :class:`CustomGate` object:

.. autoclass:: CustomGate
    :members:

In ``custom_gates`` is not given, Qiskit will attempt to use its standard-library gate objects for
the gates defined in OpenQASM 3 standard library file ``stdgates.inc``.  This sequence of gates is
available on this module, if you wish to build on top of it:

.. py:data:: STDGATES_INC_GATES

    A tuple of :class:`CustomGate` objects specifying the Qiskit constructors to use for the
    ``stdgates.inc`` include file.
"""

import functools
import warnings

from qiskit._accelerate import qasm3 as _qasm3
from qiskit.circuit import library
from qiskit.exceptions import ExperimentalWarning
from qiskit.utils import optionals as _optionals

from .experimental import ExperimentalFeatures
from .exporter import Exporter
from .exceptions import QASM3Error, QASM3ImporterError, QASM3ExporterError
from .._accelerate.qasm3 import CustomGate


STDGATES_INC_GATES = (
    CustomGate(library.PhaseGate, "p", 1, 1),
    CustomGate(library.XGate, "x", 0, 1),
    CustomGate(library.YGate, "y", 0, 1),
    CustomGate(library.ZGate, "z", 0, 1),
    CustomGate(library.HGate, "h", 0, 1),
    CustomGate(library.SGate, "s", 0, 1),
    CustomGate(library.SdgGate, "sdg", 0, 1),
    CustomGate(library.TGate, "t", 0, 1),
    CustomGate(library.TdgGate, "tdg", 0, 1),
    CustomGate(library.SXGate, "sx", 0, 1),
    CustomGate(library.RXGate, "rx", 1, 1),
    CustomGate(library.RYGate, "ry", 1, 1),
    CustomGate(library.RZGate, "rz", 1, 1),
    CustomGate(library.CXGate, "cx", 0, 2),
    CustomGate(library.CYGate, "cy", 0, 2),
    CustomGate(library.CZGate, "cz", 0, 2),
    CustomGate(library.CPhaseGate, "cp", 1, 2),
    CustomGate(library.CRXGate, "crx", 1, 2),
    CustomGate(library.CRYGate, "cry", 1, 2),
    CustomGate(library.CRZGate, "crz", 1, 2),
    CustomGate(library.CHGate, "ch", 0, 2),
    CustomGate(library.SwapGate, "swap", 0, 2),
    CustomGate(library.CCXGate, "ccx", 0, 3),
    CustomGate(library.CSwapGate, "cswap", 0, 3),
    CustomGate(library.CUGate, "cu", 4, 2),
    CustomGate(library.CXGate, "CX", 0, 2),
    CustomGate(library.PhaseGate, "phase", 1, 1),
    CustomGate(library.CPhaseGate, "cphase", 1, 2),
    CustomGate(library.IGate, "id", 0, 1),
    CustomGate(library.U1Gate, "u1", 1, 1),
    CustomGate(library.U2Gate, "u2", 2, 1),
    CustomGate(library.U3Gate, "u3", 3, 1),
)


def dumps(circuit, **kwargs) -> str:
    """Serialize a :class:`~qiskit.circuit.QuantumCircuit` object in an OpenQASM 3 string.

    Args:
        circuit (QuantumCircuit): Circuit to serialize.
        **kwargs: Arguments for the :obj:`.Exporter` constructor.

    Returns:
        str: The OpenQASM 3 serialization
    """
    return Exporter(**kwargs).dumps(circuit)


def dump(circuit, stream, **kwargs) -> None:
    """Serialize a :class:`~qiskit.circuit.QuantumCircuit` object as an OpenQASM 3 stream to
    file-like object.

    Args:
        circuit (QuantumCircuit): Circuit to serialize.
        stream (TextIOBase): stream-like object to dump the OpenQASM 3 serialization
        **kwargs: Arguments for the :obj:`.Exporter` constructor.

    """
    Exporter(**kwargs).dump(circuit, stream)


@_optionals.HAS_QASM3_IMPORT.require_in_call("loading from OpenQASM 3")
def load(filename: str):
    """Load an OpenQASM 3 program from the file ``filename``.

    Args:
        filename: the filename to load the program from.

    Returns:
        QuantumCircuit: a circuit representation of the OpenQASM 3 program.

    Raises:
        QASM3ImporterError: if the OpenQASM 3 file is invalid, or cannot be represented by a
            :class:`.QuantumCircuit`.
    """

    import qiskit_qasm3_import

    with open(filename, "r") as fptr:
        program = fptr.read()
    try:
        return qiskit_qasm3_import.parse(program)
    except qiskit_qasm3_import.ConversionError as exc:
        raise QASM3ImporterError(str(exc)) from exc


@_optionals.HAS_QASM3_IMPORT.require_in_call("loading from OpenQASM 3")
def loads(program: str):
    """Load an OpenQASM 3 program from the given string.

    Args:
        program: the OpenQASM 3 program.

    Returns:
        QuantumCircuit: a circuit representation of the OpenQASM 3 program.

    Raises:
        QASM3ImporterError: if the OpenQASM 3 file is invalid, or cannot be represented by a
            :class:`.QuantumCircuit`.
    """

    import qiskit_qasm3_import

    try:
        return qiskit_qasm3_import.parse(program)
    except qiskit_qasm3_import.ConversionError as exc:
        raise QASM3ImporterError(str(exc)) from exc


@functools.wraps(_qasm3.loads)
def loads_experimental(source, /, *, custom_gates=None, include_path=None):
    """<overridden by functools.wraps>"""
    warnings.warn(
        "This is an experimental native version of the OpenQASM 3 importer."
        " Beware that its interface might change, and it might be missing features.",
        category=ExperimentalWarning,
    )
    return _qasm3.loads(source, custom_gates=custom_gates, include_path=include_path)


@functools.wraps(_qasm3.load)
def load_experimental(pathlike_or_filelike, /, *, custom_gates=None, include_path=None):
    """<overridden by functools.wraps>"""
    warnings.warn(
        "This is an experimental native version of the OpenQASM 3 importer."
        " Beware that its interface might change, and it might be missing features.",
        category=ExperimentalWarning,
    )
    return _qasm3.load(pathlike_or_filelike, custom_gates=custom_gates, include_path=include_path)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Internal format of calibration data in target."""
from __future__ import annotations
import inspect
import warnings
from abc import ABCMeta, abstractmethod
from collections.abc import Sequence, Callable
from enum import IntEnum
from typing import Any

from qiskit.pulse.exceptions import PulseError
from qiskit.pulse.schedule import Schedule, ScheduleBlock
from qiskit.qobj.converters import QobjToInstructionConverter
from qiskit.qobj.pulse_qobj import PulseQobjInstruction
from qiskit.exceptions import QiskitError


IncompletePulseQobj = object()
"""A None-like constant that represents the PulseQobj is incomplete."""


class CalibrationPublisher(IntEnum):
    """Defines who defined schedule entry."""

    BACKEND_PROVIDER = 0
    QISKIT = 1
    EXPERIMENT_SERVICE = 2


class CalibrationEntry(metaclass=ABCMeta):
    """A metaclass of a calibration entry.

    This class defines a standard model of Qiskit pulse program that is
    agnostic to the underlying in-memory representation.

    This entry distinguishes whether this is provided by end-users or a backend
    by :attr:`.user_provided` attribute which may be provided when
    the actual calibration data is provided to the entry with by :meth:`define`.

    Note that a custom entry provided by an end-user may appear in the wire-format
    as an inline calibration, e.g. :code:`defcal` of the QASM3,
    that may update the backend instruction set architecture for execution.

    .. note::

        This and built-in subclasses are expected to be private without stable user-facing API.
        The purpose of this class is to wrap different
        in-memory pulse program representations in Qiskit, so that it can provide
        the standard data model and API which are primarily used by the transpiler ecosystem.
        It is assumed that end-users will never directly instantiate this class,
        but :class:`.Target` or :class:`.InstructionScheduleMap` internally use this data model
        to avoid implementing a complicated branching logic to
        manage different calibration data formats.

    """

    @abstractmethod
    def define(self, definition: Any, user_provided: bool):
        """Attach definition to the calibration entry.

        Args:
            definition: Definition of this entry.
            user_provided: If this entry is defined by user.
                If the flag is set, this calibration may appear in the wire format
                as an inline calibration, to override the backend instruction set architecture.
        """
        pass

    @abstractmethod
    def get_signature(self) -> inspect.Signature:
        """Return signature object associated with entry definition.

        Returns:
            Signature object.
        """
        pass

    @abstractmethod
    def get_schedule(self, *args, **kwargs) -> Schedule | ScheduleBlock:
        """Generate schedule from entry definition.

        If the pulse program is templated with :class:`.Parameter` objects,
        you can provide corresponding parameter values for this method
        to get a particular pulse program with assigned parameters.

        Args:
            args: Command parameters.
            kwargs: Command keyword parameters.

        Returns:
            Pulse schedule with assigned parameters.
        """
        pass

    @property
    @abstractmethod
    def user_provided(self) -> bool:
        """Return if this entry is user defined."""
        pass


class ScheduleDef(CalibrationEntry):
    """In-memory Qiskit Pulse representation.

    A pulse schedule must provide signature with the .parameters attribute.
    This entry can be parameterized by a Qiskit Parameter object.
    The .get_schedule method returns a parameter-assigned pulse program.

    .. see_also::
        :class:`.CalibrationEntry` for the purpose of this class.

    """

    def __init__(self, arguments: Sequence[str] | None = None):
        """Define an empty entry.

        Args:
            arguments: User provided argument names for this entry, if parameterized.

        Raises:
            PulseError: When `arguments` is not a sequence of string.
        """
        if arguments and not all(isinstance(arg, str) for arg in arguments):
            raise PulseError(f"Arguments must be name of parameters. Not {arguments}.")
        if arguments:
            arguments = list(arguments)
        self._user_arguments = arguments

        self._definition: Callable | Schedule | None = None
        self._signature: inspect.Signature | None = None
        self._user_provided: bool | None = None

    @property
    def user_provided(self) -> bool:
        return self._user_provided

    def _parse_argument(self):
        """Generate signature from program and user provided argument names."""
        # This doesn't assume multiple parameters with the same name
        # Parameters with the same name are treated identically
        all_argnames = {x.name for x in self._definition.parameters}

        if self._user_arguments:
            if set(self._user_arguments) != all_argnames:
                raise PulseError(
                    "Specified arguments don't match with schedule parameters. "
                    f"{self._user_arguments} != {self._definition.parameters}."
                )
            argnames = list(self._user_arguments)
        else:
            argnames = sorted(all_argnames)

        params = []
        for argname in argnames:
            param = inspect.Parameter(
                argname,
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            )
            params.append(param)
        signature = inspect.Signature(
            parameters=params,
            return_annotation=type(self._definition),
        )
        self._signature = signature

    def define(
        self,
        definition: Schedule | ScheduleBlock,
        user_provided: bool = True,
    ):
        self._definition = definition
        self._parse_argument()
        self._user_provided = user_provided

    def get_signature(self) -> inspect.Signature:
        return self._signature

    def get_schedule(self, *args, **kwargs) -> Schedule | ScheduleBlock:
        if not args and not kwargs:
            out = self._definition
        else:
            try:
                to_bind = self.get_signature().bind_partial(*args, **kwargs)
            except TypeError as ex:
                raise PulseError(
                    "Assigned parameter doesn't match with schedule parameters."
                ) from ex
            value_dict = {}
            for param in self._definition.parameters:
                # Schedule allows partial bind. This results in parameterized Schedule.
                try:
                    value_dict[param] = to_bind.arguments[param.name]
                except KeyError:
                    pass
            out = self._definition.assign_parameters(value_dict, inplace=False)
        if "publisher" not in out.metadata:
            if self.user_provided:
                out.metadata["publisher"] = CalibrationPublisher.QISKIT
            else:
                out.metadata["publisher"] = CalibrationPublisher.BACKEND_PROVIDER
        return out

    def __eq__(self, other):
        # This delegates equality check to Schedule or ScheduleBlock.
        if hasattr(other, "_definition"):
            return self._definition == other._definition
        return False

    def __str__(self):
        out = f"Schedule {self._definition.name}"
        params_str = ", ".join(self.get_signature().parameters.keys())
        if params_str:
            out += f"({params_str})"
        return out


class CallableDef(CalibrationEntry):
    """Python callback function that generates Qiskit Pulse program.

    A callable is inspected by the python built-in inspection module and
    provide the signature. This entry is parameterized by the function signature
    and .get_schedule method returns a non-parameterized pulse program
    by consuming the provided arguments and keyword arguments.

    .. see_also::
        :class:`.CalibrationEntry` for the purpose of this class.

    """

    def __init__(self):
        """Define an empty entry."""
        self._definition = None
        self._signature = None
        self._user_provided = None

    @property
    def user_provided(self) -> bool:
        return self._user_provided

    def define(
        self,
        definition: Callable,
        user_provided: bool = True,
    ):
        self._definition = definition
        self._signature = inspect.signature(definition)
        self._user_provided = user_provided

    def get_signature(self) -> inspect.Signature:
        return self._signature

    def get_schedule(self, *args, **kwargs) -> Schedule | ScheduleBlock:
        try:
            # Python function doesn't allow partial bind, but default value can exist.
            to_bind = self._signature.bind(*args, **kwargs)
            to_bind.apply_defaults()
        except TypeError as ex:
            raise PulseError("Assigned parameter doesn't match with function signature.") from ex
        out = self._definition(**to_bind.arguments)
        if "publisher" not in out.metadata:
            if self.user_provided:
                out.metadata["publisher"] = CalibrationPublisher.QISKIT
            else:
                out.metadata["publisher"] = CalibrationPublisher.BACKEND_PROVIDER
        return out

    def __eq__(self, other):
        # We cannot evaluate function equality without parsing python AST.
        # This simply compares weather they are the same object.
        if hasattr(other, "_definition"):
            return self._definition == other._definition
        return False

    def __str__(self):
        params_str = ", ".join(self.get_signature().parameters.keys())
        return f"Callable {self._definition.__name__}({params_str})"


class PulseQobjDef(ScheduleDef):
    """Qobj JSON serialized format instruction sequence.

    A JSON serialized program can be converted into Qiskit Pulse program with
    the provided qobj converter. Because the Qobj JSON doesn't provide signature,
    conversion process occurs when the signature is requested for the first time
    and the generated pulse program is cached for performance.

    .. see_also::
        :class:`.CalibrationEntry` for the purpose of this class.

    """

    def __init__(
        self,
        arguments: Sequence[str] | None = None,
        converter: QobjToInstructionConverter | None = None,
        name: str | None = None,
    ):
        """Define an empty entry.

        Args:
            arguments: User provided argument names for this entry, if parameterized.
            converter: Optional. Qobj to Qiskit converter.
            name: Name of schedule.
        """
        super().__init__(arguments=arguments)

        self._converter = converter or QobjToInstructionConverter(pulse_library=[])
        self._name = name
        self._source: list[PulseQobjInstruction] | None = None

    def _build_schedule(self):
        """Build pulse schedule from cmd-def sequence."""
        with warnings.catch_warnings():
            warnings.simplefilter(action="ignore", category=DeprecationWarning)
            # `Schedule` is being deprecated in Qiskit 1.3
            schedule = Schedule(name=self._name)
        try:
            for qobj_inst in self._source:
                for qiskit_inst in self._converter._get_sequences(qobj_inst):
                    schedule.insert(qobj_inst.t0, qiskit_inst, inplace=True)
            self._definition = schedule
            self._parse_argument()
        except QiskitError as ex:
            # When the play waveform data is missing in pulse_lib we cannot build schedule.
            # Instead of raising an error, get_schedule should return None.
            warnings.warn(
                f"Pulse calibration cannot be built and the entry is ignored: {ex.message}.",
                UserWarning,
            )
            self._definition = IncompletePulseQobj

    def define(
        self,
        definition: list[PulseQobjInstruction],
        user_provided: bool = False,
    ):
        # This doesn't generate signature immediately, because of lazy schedule build.
        self._source = definition
        self._user_provided = user_provided

    def get_signature(self) -> inspect.Signature:
        if self._definition is None:
            self._build_schedule()
        return super().get_signature()

    def get_schedule(self, *args, **kwargs) -> Schedule | ScheduleBlock | None:
        if self._definition is None:
            self._build_schedule()
        if self._definition is IncompletePulseQobj:
            return None
        return super().get_schedule(*args, **kwargs)

    def __eq__(self, other):
        if isinstance(other, PulseQobjDef):
            # If both objects are Qobj just check Qobj equality.
            return self._source == other._source
        if isinstance(other, ScheduleDef) and self._definition is None:
            # To compare with other schedule def, this also generates schedule object from qobj.
            self._build_schedule()
        if hasattr(other, "_definition"):
            return self._definition == other._definition
        return False

    def __str__(self):
        if self._definition is None:
            # Avoid parsing schedule for pretty print.
            return "PulseQobj"
        if self._definition is IncompletePulseQobj:
            return "None"
        return super().__str__()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test the piecewise polynomial Pauli rotations."""

import unittest
from collections import defaultdict
import numpy as np
from ddt import ddt, data, unpack

from qiskit.circuit import QuantumCircuit
from qiskit.circuit.library.arithmetic.piecewise_polynomial_pauli_rotations import (
    PiecewisePolynomialPauliRotations,
)
from qiskit.quantum_info import Statevector
from test import QiskitTestCase  # pylint: disable=wrong-import-order


@ddt
class TestPiecewisePolynomialRotations(QiskitTestCase):
    """Test the piecewise polynomial Pauli rotations."""

    def assertFunctionIsCorrect(self, function_circuit, reference):
        """Assert that ``function_circuit`` implements the reference function ``reference``."""
        num_state_qubits = function_circuit.num_state_qubits
        num_ancilla_qubits = function_circuit.num_ancillas
        circuit = QuantumCircuit(num_state_qubits + 1 + num_ancilla_qubits)
        circuit.h(list(range(num_state_qubits)))
        circuit.append(function_circuit.to_instruction(), list(range(circuit.num_qubits)))

        statevector = Statevector(circuit)

        probabilities = defaultdict(float)
        for i, statevector_amplitude in enumerate(statevector):
            i = bin(i)[2:].zfill(circuit.num_qubits)[num_ancilla_qubits:]
            probabilities[i] += np.real(np.abs(statevector_amplitude) ** 2)

        unrolled_probabilities = []
        unrolled_expectations = []
        for i, probability in probabilities.items():
            x, last_qubit = int(i[1:], 2), i[0]
            if last_qubit == "0":
                expected_amplitude = np.cos(reference(x)) / np.sqrt(2**num_state_qubits)
            else:
                expected_amplitude = np.sin(reference(x)) / np.sqrt(2**num_state_qubits)

            unrolled_probabilities += [probability]
            unrolled_expectations += [np.real(np.abs(expected_amplitude) ** 2)]

        np.testing.assert_almost_equal(unrolled_probabilities, unrolled_expectations)

    @data(
        (1, [0], [[1]]),
        (2, [0, 2], [[2], [-0.5, 1]]),
        (3, [0, 2, 5], [[1, 0, -1], [2, 1], [1, 1, 1]]),
        (4, [2, 5, 7, 16], [[1, -1], [1, 2, 3], [1, 2, 3, 4]]),
        (3, [0, 1], [[1, 0], [1, -2]]),
    )
    @unpack
    def test_piecewise_polynomial_function(self, num_state_qubits, breakpoints, coeffs):
        """Test the piecewise linear rotations."""

        def pw_poly(x):
            for i, point in enumerate(reversed(breakpoints[: len(coeffs)])):
                if x >= point:
                    # Rescale the coefficients to take into account the 2 * theta argument from the
                    # rotation gates
                    rescaled_c = [coeff / 2 for coeff in coeffs[-(i + 1)][::-1]]
                    return np.poly1d(rescaled_c)(x)
            return 0

        pw_polynomial_rotations = PiecewisePolynomialPauliRotations(
            num_state_qubits, breakpoints, coeffs
        )

        self.assertFunctionIsCorrect(pw_polynomial_rotations, pw_poly)

    def test_piecewise_polynomial_rotations_mutability(self):
        """Test the mutability of the linear rotations circuit."""

        def pw_poly(x):
            for i, point in enumerate(reversed(breakpoints[: len(coeffs)])):
                if x >= point:
                    # Rescale the coefficients to take into account the 2 * theta argument from the
                    # rotation gates
                    rescaled_c = [coeff / 2 for coeff in coeffs[-(i + 1)][::-1]]
                    return np.poly1d(rescaled_c)(x)
            return 0

        pw_polynomial_rotations = PiecewisePolynomialPauliRotations()

        with self.subTest(msg="missing number of state qubits"):
            with self.assertRaises(AttributeError):  # no state qubits set
                _ = str(pw_polynomial_rotations.draw())

        with self.subTest(msg="default setup, just setting number of state qubits"):
            pw_polynomial_rotations.num_state_qubits = 2
            self.assertFunctionIsCorrect(pw_polynomial_rotations, lambda x: 1 / 2)

        with self.subTest(msg="setting non-default values"):
            breakpoints = [0, 2]
            coeffs = [[0, -2 * 1.2], [-2 * 1, 2 * 1, 2 * 3]]
            pw_polynomial_rotations.breakpoints = breakpoints
            pw_polynomial_rotations.coeffs = coeffs
            self.assertFunctionIsCorrect(pw_polynomial_rotations, pw_poly)

        with self.subTest(msg="changing all values"):
            pw_polynomial_rotations.num_state_qubits = 4
            breakpoints = [1, 3, 6]
            coeffs = [[0, -2 * 1.2], [-2 * 1, 2 * 1, 2 * 3], [-2 * 2]]
            pw_polynomial_rotations.breakpoints = breakpoints
            pw_polynomial_rotations.coeffs = coeffs
            self.assertFunctionIsCorrect(pw_polynomial_rotations, pw_poly)


if __name__ == "__main__":
    unittest.main()
